#include <stdio.h>
#include <string.h>

int limit  = 32768;
int recurse= 0;
int part   = 0;
int partlen= 0;
FILE *output = stdout;

void ProcessFile (FILE *output, FILE *input)
{
  char data[512], *p;
  FILE *newfile;
  int count;
  
  while (!feof (input)) {
    if ((p = fgets (data, 512, input)) == NULL)
      break;

    if (strchr (data, '[') && strchr (data, ']')) {
      if (2*partlen >= limit) {
	fprintf (output, "\\n\\\n\";\n\nstatic char * _CompiledResource_%d = \"\\\n", part++);
	partlen = 0;
      }
    }

    if (strchr (data, '[') && strchr (data, ']') &&
	strstr (data, "Resource")) {
    skipcomment:
      if ((p = fgets (data, 512, input)) == NULL)
	break;
      while (*p == ' ' || *p == '\t')
	p++;
      if (*p == '#')
	goto skipcomment;
      if (strrchr (p, '\n'))
	*(strrchr (p, '\n')) = '\0';

      if (!recurse)
	continue;

      if ((newfile = fopen (p, "r")) == NULL)
	continue;

      ProcessFile (output, newfile);
      fclose (newfile);
      continue;
    }

    while (*p) {
      switch (*p) {
      case '\\':
	fputc ('\\', output); partlen += 2;
	fputc ('\\', output);
	break;
      case '\"':
	fputc ('\\', output); partlen += 2;
	fputc ('\"', output);
	break;
      case '\n':
	fputc ('\\', output); partlen += 4;
	fputc ('n',  output);
	fputc ('\\', output);
	fputc ('\n', output);
	break;
      default:
	fputc (*p, output);   partlen++;
	break;
      }
      p++;
    }
  }
}

int main (int argc, char *argv[])
{
  int count=1, sfpar, fnpar;
  FILE *input;

  if (argc < 2) {
  usage:
    fprintf (stderr, "Benutzung:\n");
    fprintf (stderr, "\trescomp [-l limit] [-r] [-n <funcname>] <sourcefiles>\n\n");
    fprintf (stderr, "\tKompiliert die Resourcedatei <source> in einen C-Quelltext nach\n");
    fprintf (stderr, "\tstdout. Dieser Quelltext exportiert die parameter- und ergebnis-\n");
    fprintf (stderr, "\tlose Funktion <funcname> (Default: \"LoadInternalRes\"), die, vom\n");
    fprintf (stderr, "\tHauptmodul aus aufgerufen, alle in <source> deklarierten Resour-\n");
    fprintf (stderr, "\tcen laedt.\n\n");
    fprintf (stderr, "\t-l\tSetzt das Limit in Bytes fuer die Maximale Groesse\n");
    fprintf (stderr, "\t\teines Strings (Compilerabhaengig). Kann mit dem Suffix\n");
    fprintf (stderr, "\t\t'k' auf Kilobyte skaliert werden. \"-l 32k\" ist Default\n");
    fprintf (stderr, "\t-r\tWeist das Programm an, auch referenzierte Resourceda-\n");
    fprintf (stderr, "\t\tteien zu durchsuchen.\n\n");
    return 1;
  }
  while (*argv[count] == '-') {
    switch (argv[count][1]) {
    case 'l':
      limit = atoi (argv[++count]);
      if (strchr (argv[count], 'k'))
	limit *= 1024;

      if (limit < 1024) {
	fprintf (stderr, "Fehler: String-Limit zu klein. Minimal 1k.\n\n");
	goto usage;
      }
      break;
    case 'r':
      recurse = 1;
      break;
    case 'n':
      fnpar = ++count;
      break;
    default:
      goto usage;
    }
    count++;
  }

  if (argc <= count) {
    fprintf (stderr, "Error! No filename specified\n\n");
    goto usage;
  }

  fprintf (output, "\
\
#include <curses.h>\n\
\n\
#ifdef NOPANEL\n\
#include <nopanel.h>\n\
#else\n\
#include <panel.h>\n\
#endif\n\
\n\
#include <resource.h>\n\
\n");

  for (sfpar = count; sfpar < argc; sfpar++) {
    if ((input = fopen (argv[sfpar], "r")) == NULL) {
      fprintf (stderr, "Error! Cannot open input file %s (ignored)\n", argv[sfpar]);
      continue;
    }
    fprintf     (output, "static char * _CompiledResource_%d = \"\\\n", part++);
    ProcessFile (output, input);
    fprintf     (output, "\\n\\\n\";\n\n");

    fclose (input);
  }

  fprintf (output, "void %s (void)\n", (argc<=fnpar)?"LoadInternalRes":argv[fnpar]);
  fprintf (output, "{\n");

  for (count=0; count < part; count++)
    fprintf (output, "  AddMoreResources (LoadResource (NULL, _CompiledResource_%d));\n", count);

  fprintf (output, "}\n");

  if (output != stdout)
    fclose (output);

  return 0;
}
