/*****************************************************************************
******************************************************************************
*****                                                                    *****
*****     resource.c                                                     *****
*****                   Bildschirmfunktionen                             *****
*****                                                                    *****
*****     ??.03.93      erstellt                                   FP    *****
*****                                                                    *****
******************************************************************************
*****************************************************************************/

#include <curses.h>

#ifdef NOPANEL
#include <nopanel.h>
#else
#include <panel.h>
#endif

#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <memory.h>
#include <time.h>
#include <sys/socket.h>
#include "resource.h"

#ifdef TERMIO
#  include <termio.h>
#endif

RESOURCE *mainres = NULL;

static char *		HelpStack[30];
static int 		HelpCount = -1;

static int		KeyBuffer[100];
static int		KeyHooks[30];
static int		KeyAlias[30][2];
static int	       *KeyMacros[30];
static void		(*KH_Funcs[30]) (void);
static int		KB_Count = 0;
static int		KH_Count = 0;
static int		KA_Count = 0;
static int		KM_Count = 0;
static int		isBreak;
static int		BlockBreak;
static FINFO *		AbbruchMeld = NULL;
static char 		linebuffer[256];
static int		lbindex  = -1;
static int		monoflag = 0;

#ifdef NEWTERM
static SCREEN *		actterm  = NULL;
#else
static void *		actterm  = NULL;
#endif

extern int		getcval			(char *);
extern void             LoadExternResource      (FILE *, char **, char *, char *);

#ifdef SIGACTION
static struct sigaction act;
#endif

/*****************************************************************************
***** Strutkur fuer die Anzeige von Texten                               *****
*****************************************************************************/

struct _lineinfo {
  int length;
  int type;
  char *ptr;
};

/*****************************************************************************
***** Lokale Funktionen                                                  *****
*****************************************************************************/

char accpref[10] = { 0 };

int                     str2val                 (char *, char **);

static TEXT *		LoadTextResource 	(FILE *, char **, char *, char *, char *);
static SBOX *		LoadSboxResource 	(FILE *, char **, char *, char *);
static MENU *		LoadMenuResource 	(FILE *, char **, char *, char *);
static MESSAGE *	LoadMsgResource 	(FILE *, char **, char *, char *);
static FORM *		LoadFormResource 	(FILE *, char **, char *, char *);
static TVECTOR *	LoadTvecResource	(FILE *, char **, char *, char *);

static RESOURCE *	AddResource 		(RESOURCE *, RESOURCE *);
static RESOURCE *	MergeTrees 		(RESOURCE *, RESOURCE *);

static void 		KillSboxResource 	(SBOX *);
static void 		KillMenuResource 	(MENU *);
static void 		KillTextResource 	(TEXT *);
static void 		KillMsgResource 	(MESSAGE *);
static void 		KillFormResource 	(FORM *);
static void		KillTvecResource	(TVECTOR *);

static char *		_getlinebuffer		(FILE *, char **);
static void 		_getinfo 		(struct _lineinfo *, char *, int, int);

static void		InitTerm		(void);
static void		FinishTerm		(void);
static void		SigSetBreak		(int);

/*
 * Mini-Taschenrechner, damit in der Eingabedatei Rechnungen ausge-
 * fuehrt werden duerfen. Einleitung der Rechnung mit %, Formeln in
 * Lisp-Notation, z.B. %(+ 100 100), %(+ 100 (* 200 5))
 * Konstanten sind von Dollarzeichen eingeschlossen, z.B, $PAGELENGTH$
 **/

int str2val (char *input, char **next)
{
  int result, brack;
  char *iter=input;

  if (input == NULL) {
    *next = NULL;
    return -1;
  }
  while (*iter == ' ' || *iter == '\t')
    iter++;

  if (*iter == '$') {
    result = getcval (iter++);

    while (*iter && *iter != '$')
      iter++;

    if (*iter)
      iter++;
  }
  else if (*iter == '(') {
    result = minical (iter);
    brack= 1; iter++;

    while (brack) {
      if (*iter == '\0')
	break;
      else if (*iter == '(')
	brack++;
      else if (*iter == ')')
	brack--;
      iter++;
    }
  }
  else if (isdigit (*iter)) {
    result = atoi (iter);
    while (isdigit (*iter))
      iter++;
  }
  while (*iter == ' ' || *iter == '\t')
    iter++;

  *next = iter;
  return result;
}

int minical (char *input)
{
  int result, val1, val2;
  char *iter=input, op;

  if (*iter++ != '(')
    return -1;
  while (*iter == ' ' || *iter == '\t')
    iter++;

  op = *iter++;

  val1 = str2val (iter, &iter);
  val2 = str2val (iter, &iter);

  switch (op) {
  case '+': result = val1 + val2; break;
  case '*': result = val1 * val2; break;
  case '-': result = val1 - val2; break;
  case '/': result = val1 / val2; break;
  default:  result = -1;          break;
  }
  return result;
}

char *calcstr (char *in, char *out)
{
  static char str[256];
  char *temp = out;
  int count, brack;

  while (*in) {
    if (*in == '%' || *in == '&') {
      str[0] = *++in; if (*in) in++;
      brack  = 1;  count = 1;
      
      while (brack) {
	if (*in == '\0' || *in == '\n')
	  break;
	
	if (*in == '(')
	  brack++;
	else if (*in == ')')
	  brack--;
	
	str[count++] = *in++;
      }
      str[count] = '\0';

      out += sprintf (out, "%d", minical (str));
    }
    else
      *out++ = *in++;
  }
  *out = '\0';
  
  return temp;
}

/*
 * Lese ein Zeichen aus der Resourcedatei
 * Wird ein Kommentarzeichen '#' gelesen, so wird der Rest der Zeile
 * ignoriert, das abschliessende CR aber zurueckgegeben
 **/

char *FGETSC (FILE *datei, char **ptr, char *target)
{
  char *index = target;
  int count=0;

  if (FEOFC (datei, ptr))
    return NULL;

  while (count<254) {
    *index = FGETCC (datei, ptr);
    if (FEOFC (datei, ptr))
      break;

    if (*index == '\x0a' || *index == '\x0d') {
      if (*index == '\x0d')
	if ((*index = FGETCC (datei, ptr)) != '\x0a')
	  UNGETCC (*index, datei, ptr);
      break;
    }
    count++;
    index++;
  }
  *index++ = '\n';
  *index   = '\0';

  return target;
}

char *_getlinebuffer (FILE *datei, char **ptr)
{
  char *c, *cos, *cod;
  int count, apc, esc=0;

 gnlnl:
  if ((c = FGETSC (datei, ptr, linebuffer)) == NULL)
    return NULL;

 glcck:
  while (*c && (*c == ' ' || *c == '\t'))
    c++;

  if (*c == '#' || *c == '\n' || *c == '\0')
    goto gnlnl;

  if (*c == '^') {
    for (apc=1; isalpha(c[apc]); apc++) {
      for (count=0; accpref[count]; count++) {
	if (c[apc] == accpref[count])
	  break;
      }
      if (c[apc] == accpref[count])
	break;
    }
    if (c[apc] == '\n' || c[apc] == '\0' || c[apc] != accpref[count])
      goto gnlnl;

    cod = c;
    cos = c+1;

    while (isalpha (*cos))
      cos++;

    while (*cos && (*cos == ' ' || *cos == '\t'))
      cos++;

    do {
      *cod++ = *cos;
    }
    while (*cos++);

    goto glcck;
  }

  if (linebuffer[0] == '[' && linebuffer[1] == '^')
    c = linebuffer+2;

  while (*c && *c != '\n') {
    if (*c == '#') {
      *c++ = '\n';
      *c   = '\0';
      break;
    }
    if (*c == '^' && !esc) {
      for (apc=1; isalpha(c[apc]); apc++) {
	for (count=0; accpref[count]; count++) {
	  if (c[apc] == accpref[count])
	    break;
	}
	if (c[apc] == accpref[count])
	  break;
      }
      if (c[apc] == '\n' || c[apc] == '\0' || c[apc] != accpref[count]) {
	*c++ = '\n';
	*c   = '\0';
	break;
      }

      cod = c;
      cos = c+1;

      while (isalpha(*cos))
	cos++;

      while (*cos && (*cos == ' ' || *cos == '\t'))
	cos++;

      do {
	*cod++ = *cos;
      }
      while (*cos++);
    }
    else if (*c == '\\' && !esc) {
      esc=1;
      c++;
    }
    else {
      esc=0;
      c++;
    }
  }

  lbindex = 0;
  return linebuffer;
}

char freadc (FILE *datei, char **ptr)
{
  char *c;

  if (lbindex == -1 || linebuffer[lbindex] == '\0')
    if (_getlinebuffer (datei, ptr) == NULL)
      return '\0';

  return linebuffer[lbindex++];
}

void fungetc (int c)
{
  if (lbindex)
    linebuffer[--lbindex] = c;
}

/*****************************************************************************
***** freadi:      Einlesen einer Zahl aus einem Stream                  *****
*****              Vor und nach der Zahl duerfen beliebig viele Spaces / *****
*****              Tabs stehen; bei der ersten Nicht-Ziffer wird abgebro-*****
*****              chen. Konnte keine Ziffer gelesen werden, wird -1 zu- *****
*****              rueckgegeben und der Stream bleibt unveraendert       *****
*****************************************************************************/

int freadi (FILE *datei, char **ptr)
{
  int result=0, brack=1, mflag=0, count=0;
  char temp[256];

  if (lbindex == -1 || linebuffer[lbindex] == '\n' || linebuffer[lbindex] == '\0')
  newline:
    if (_getlinebuffer (datei, ptr) == NULL)
      return -1;

  while (linebuffer[lbindex] == ' ' || linebuffer[lbindex] == '\t')
    lbindex++;
  if (linebuffer[lbindex] == '\n' || linebuffer[lbindex] == '\0')
    goto newline;

  if (linebuffer[lbindex] == '-' && isdigit(linebuffer[lbindex+1])) {
    mflag = 1;
    lbindex++;
  }
  if (linebuffer[lbindex] == '%' && linebuffer[lbindex+1] == '(') {
    temp[count++] = linebuffer[++lbindex]; lbindex++;
    
    while (brack) {
      if (linebuffer[lbindex] == '\n' || linebuffer[lbindex] == '\0')
	return -1;
      if (linebuffer[lbindex] == '(')
	brack++;
      else if (linebuffer[lbindex] == ')')
	brack--;
      temp[count++] = linebuffer[lbindex++];
    }
    temp[count] = '\0';
    
    result = minical (temp);
  }
  else if (isdigit(linebuffer[lbindex])) {
    while (isdigit(linebuffer[lbindex]))
      result = 10 * result + (int) (linebuffer[lbindex++] - '0');
  }
  else
    return -1;

  while (linebuffer[lbindex] == ' ' && linebuffer[lbindex] == '\t')
    lbindex++;
  if (linebuffer[lbindex] == '\n')
    lbindex++;

  if (mflag)
    result = -result;

  return result;
}

/****************************************************************************
*****  freads  -   Liest einen String aus Stream. Der Speicherbereich   *****
*****              wird dynamisch allokiert.                            *****
*****              Vor und nach dem String werden Spaces / Tabs und     *****
*****              genau ein Zeilenumbruch ignoriert.                   *****
*****************************************************************************/

char *freads (FILE *datei, char **ptr)
{
  static char temp[256], calc[256];
  int index=0, count, brack, eat=0;
  char *result;

  if (lbindex == -1 || linebuffer[lbindex] == '\n' || linebuffer[lbindex] == '\0')
  newline:
    if (_getlinebuffer (datei, ptr) == NULL)
      return NULL;

  while (linebuffer[lbindex] == ' ' || linebuffer[lbindex] == '\t')
    lbindex++;
  if (linebuffer[lbindex] == '\n' || linebuffer[lbindex] == '\0')
    goto newline;


  do {
    if ((linebuffer[lbindex] == '[' && lbindex       ==  0 ) ||
	(linebuffer[lbindex] == ']' && linebuffer[0] == '[') ||
	(linebuffer[lbindex] == ':')  ||
	(linebuffer[lbindex] == '\n') ||
	(linebuffer[lbindex] == '\0'))
      break;
    if (linebuffer[lbindex] == '%' && linebuffer[lbindex+1] == '(') {
      brack = 1; count = 1; lbindex++;
      calc[0] = linebuffer[lbindex++];

      while (brack) {
	if (linebuffer[lbindex] == '\n' || linebuffer[lbindex] == '\0')
	  break;
	if (linebuffer[lbindex] == '(')
	  brack++;
	else if (linebuffer[lbindex] == ')')
	  brack--;
	calc[count++] = linebuffer[lbindex++];
      }
      calc[count] = '\0';
      
      if (brack)
	temp[index++] = '0';
      else
	index += sprintf (temp + index, "%d", minical (calc));
    }
    else if (linebuffer[lbindex] == '\\') {
      if (lbindex++ == 0)
	eat = 1;

      if (linebuffer[lbindex] == '\n' || linebuffer[lbindex] == '\0') {
	if (_getlinebuffer (datei, ptr) == NULL)
	  linebuffer[lbindex=0] = '\0';
      }
      else {
	switch (linebuffer[lbindex]) {
	case 'n': temp[index] = '\n'; break;
	case 'r': temp[index] = '\r'; break;
	case 't': temp[index] = '\t'; break;
	default:  temp[index] = linebuffer[lbindex]; break;
	}
	index++;
	lbindex++;
      }
    }
    else
      temp[index++] = linebuffer[lbindex++];
  }
  while (TRUE);

  if (linebuffer[lbindex] == '\n')
    lbindex++;

  while (index>0 && (temp[index-1] == ' ' || temp[index-1] == '\t'))
    index--;

  if ((temp[0] == '@' && !eat) || (index==0 && linebuffer[lbindex]==':'))
    return NULL;

  temp[index] = '\0';
  
  if ((result = (char *) malloc ((index+1) * sizeof (char))) == NULL)
    return NULL;

  return (char *) memcpy (result, temp, (index+1) * sizeof (char));
}

/****************************************************************************
***** Ein paar Hilfsfunktionen, die die Arbeit mit Forms erleichtern    *****
*****************************************************************************/

char *ShortString (char *input)
{
  char *ptr1 = input, *ptr2 = input, *ptr3 = input;


  ptr3 += strlen(input);

  while (ptr3>input && (*(ptr3-1)==' ' || *(ptr3-1)=='\n'))
    *--ptr3 = '\0';

  while (*ptr1 == ' ')
    ptr1++;

  while (*ptr1 != '\0')
    *ptr2++ = *ptr1++;

  while (ptr2 != ptr1)
    *ptr2++ = '\0';

  return input;
}

int IsStringEmpty (char *input)
{
  while (*input == ' ')
    input++;

  if (*input == '\0')
    return 1;

  return 0;
}
     
/****************************************************************************
***** Loader  -    Die folgenden Funktionen dienen dem Laden der        *****
*****              Resourcen aus einer Datei                            *****
*****************************************************************************/

TEXT *LoadTextResource (FILE *datei, char **ptr, char *fname, char *rname, char *HelpContext)
{
  long length = 0L;
  TEXT *result;
  int  eat=0;
  char c=0;

  if ((result = (TEXT *) malloc (sizeof (TEXT))) == NULL)
    return NULL;
  if ((result->title = freads (datei, ptr)) == NULL) {
    free (result);
    return NULL;
  }
  result->rname   = rname;

  if (datei) {
    result->offset  = ftell  (datei);
    result->filename= strdup (fname);
  }
  else {
    result->offset  = -1;
    result->filename= *ptr;
  }

  while ((c != '[' || eat) && !FEOFC (datei, ptr)) {
    if ((c = freadc (datei, ptr)) == '\\') {
      c = freadc (datei, ptr);
      length += 2;
      eat = 1;
    }
    else {
      length++;
      eat = 0;
    }
  }
  if (c == '[')
    fungetc (c);

  result->length      = length - 1;
  result->HelpContext = HelpContext;
  
  return result;
}

SBOX *LoadSboxResource (FILE *datei, char **ptr, char *rname, char *HelpContext)
{
  int index, anzahl;
  SBOX *result;

  if ((result = (SBOX *) malloc (sizeof (SBOX))) == NULL)
    return NULL;

  result->rname = rname;
  result->anzahl= anzahl = freadi (datei, ptr);
  result->ypos  = freadi (datei, ptr);
  result->xpos  = freadi (datei, ptr);
  result->title = freads (datei, ptr);

  if ((result->resid = (int *) malloc (anzahl * sizeof (int))) == NULL) {
    free (result->title);
    free (result);
    return NULL;
  }
  if ((result->items = (char **) malloc (anzahl * sizeof (char *))) == NULL) {
    free (result->resid);
    free (result->title);
    free (result);
    return NULL;
  }

  if (result->title != NULL)
    result->xsize = strlen (result->title);
  else
    result->xsize = 0;
  
  for (index=0; index < anzahl; index++) {
    result->resid[index] = freadi (datei, ptr);
    result->items[index] = freads (datei, ptr);
    if (result->items[index] == NULL)
      continue;
    else if (strlen (result->items[index]) > result->xsize)
      result->xsize = strlen (result->items[index]);
  }

  result->ysize       = anzahl;
  result->HelpContext = HelpContext;
  
  return result;
}

MENU *LoadMenuResource (FILE *datei, char **ptr, char *rname, char *HelpContext)
{
  int index, c, anzahl;
  MENU *result;

  if ((result = (MENU *) malloc (sizeof (MENU))) == NULL)
    return NULL;

  result->rname  = rname;
  result->anzahl = freadi (datei, ptr);

  for (index=0; index<result->anzahl; index++) {
    result->ypos[index] = freadi (datei, ptr);
    result->xpos[index] = freadi (datei, ptr);
    result->resid[index]= freadi (datei, ptr);
    result->text[index] = freads (datei, ptr);

    c = freadc (datei, ptr);
    if (c == ':')
      result->submenu[index] = freads (datei, ptr);
    else {
      result->submenu[index] = NULL;
      fungetc (c);
    }
  }

  result->HelpContext = HelpContext;

  return result;
}

MESSAGE *LoadMsgResource (FILE *datei, char **ptr, char *rname, char *HelpContext)
{
  MESSAGE *result;

  if ((result = (MESSAGE *) malloc (sizeof (MESSAGE))) == NULL) 
    return NULL;

  result->rname       = rname;
  result->HelpContext = HelpContext;

  if ((result->text = freads (datei, ptr)) == NULL) {
    free (result);
    return NULL;
  }
  return result;
}

FORM *LoadFormResource (FILE *datei, char **ptr, char *rname, char *HelpContext)
{
  char *gettype, *tmp, def[40];
  int index=-1, c, anzahl, di;
  FORM *result;

  if ((result = (FORM *) malloc (sizeof (FORM))) == NULL)
    return NULL;

  result->rname = rname;
  result->anzahl= anzahl = freadi (datei, ptr);
  result->yres  = freadi (datei, ptr);
  result->xres  = freadi (datei, ptr);
  result->ypos  = freadi (datei, ptr);
  result->xpos  = freadi (datei, ptr);
  result->title = freads (datei, ptr);
  result->field = 0;
  result->HelpContext = HelpContext;

  if ((result->lypos = (int *) malloc (anzahl * sizeof (int))) == NULL) {
    free (result);
    return NULL;
  }
  if ((result->lxpos = (int *) malloc (anzahl * sizeof (int))) == NULL) {
    free (result->lypos);
    free (result);
    return NULL;
  }
  if ((result->fypos = (int *) malloc (anzahl * sizeof (int))) == NULL) {
    free (result->lxpos); free (result->lypos);
    free (result);
    return NULL;
  }
  if ((result->fxpos = (int *) malloc (anzahl * sizeof (int))) == NULL) {
    free (result->lxpos); free (result->lypos);
    free (result->fypos);
    free (result);
    return NULL;
  }
  if ((result->flen = (int *) malloc (anzahl * sizeof (int))) == NULL) {
    free (result->lxpos); free (result->lypos);
    free (result->fxpos); free (result->fypos);
    free (result);
    return NULL;
  }
  if ((result->resid = (int *) malloc (anzahl * sizeof (int))) == NULL) {
    free (result->lxpos); free (result->lypos);
    free (result->fxpos); free (result->fypos);
    free (result->flen);
    free (result);
    return NULL;
  }
  if ((result->ftype = (int *) malloc (anzahl * sizeof (int))) == NULL) {
    free (result->lxpos); free (result->lypos);
    free (result->fxpos); free (result->fypos);
    free (result->resid); free (result->flen);
    free (result);
    return NULL;
  }
  if ((result->label = (char **) malloc (anzahl * sizeof (char *))) == NULL) {
    free (result->lxpos); free (result->lypos);
    free (result->fxpos); free (result->fypos);
    free (result->resid); free (result->flen);
    free (result->ftype);
    free (result);
    return NULL;
  }
  if ((result->stand = (char **) malloc (anzahl * sizeof (char *))) == NULL) {
    free (result->lxpos); free (result->lypos);
    free (result->fxpos); free (result->fypos);
    free (result->resid); free (result->flen);
    free (result->ftype); free (result->label);
    free (result);
    return NULL;
  }

  result->buffer = NULL;

  for (index=0; index<anzahl; index++) {
    result->lypos[index] = freadi (datei, ptr);
    result->lxpos[index] = freadi (datei, ptr);
    result->label[index] = freads (datei, ptr);
    result->stand[index] = NULL;

    c = freadc (datei, ptr);

    if (result->lypos[index] == -1 && result->label[index] == NULL && c != ':') {
      fungetc (c);
      break;
    }

    if (c == ':') {
      result->resid[index] = freadi (datei, ptr);
      result->fypos[index] = freadi (datei, ptr);
      result->fxpos[index] = freadi (datei, ptr);
      result->flen [index] = freadi (datei, ptr);
      result->ftype[index] = di = 0;
      gettype              = freads (datei, ptr);

      if (gettype != NULL) {
	tmp = gettype;
	while (*tmp) {
	  switch (toupper (*tmp)) {
	  case 'A':
	    result->ftype[index] |= IT_ALPHA;
	    break;
	  case 'N':
	    result->ftype[index] |= IT_NUMERIC;
	    break;
	  case 'C':
	    result->ftype[index] |= IT_CHAR;
	    break;
	  case 'U':
	    result->ftype[index] |= IT_UPCASE;
	    break;
	  case 'D':
	    result->ftype[index] |= IT_DCASE;
	    break;
	  case 'S':
	    result->ftype[index] |= IT_SPACE;
	    break;
	  case 'F':
	    result->ftype[index] |= IT_FKEYEXIT;
	    break;
	  case '-':
	    result->ftype[index] |= IT_INACTIVE;
	    break;
	  case '?':
	    result->ftype[index]  = IT_YESNO;
	    break;
	  case 'L':
	    result->ftype[index] |= IT_DOWNLIGHT;
	    break;
	  case '!': tmp++;
	    if (*tmp == 'E')
	      result->ftype[index] |= IT_ACTENTER;
	    else if (*tmp == 'L')
	      result->ftype[index] |= IT_ACTLEAVE;
	    break;
	  case '\'': tmp++;
	    while (*tmp != '\'')
	      def[di++] = *tmp++;
	    break;
	  }
	  tmp++;
	}
	free (gettype);
      }

      if (result->ftype[index] == 0)
	result->ftype[index] = IT_ALNUM;
      if (di) {
	def[di] = '\0';
	result->stand[index] = strdup (def);
      }
    }
    else {
      result->fypos[index] = 0;
      result->fxpos[index] = 0;
      result->flen [index] = 0;
      fungetc (c);
    }
  }

  if (result->title != NULL && strlen (result->title) > result->xres)
    result->xres = strlen (result->title);

  return result;
}

TVECTOR *LoadTvecResource (FILE *datei, char **ptr, char *rname, char *HelpContext)
{
  TVECTOR *result;
  int index;

  if ((result = (TVECTOR *) malloc (sizeof (TVECTOR))) == NULL)
    return NULL;

  result->rname       = rname;
  result->HelpContext = HelpContext;
  result->anzahl      = freadi (datei, ptr);

  if ((result->elements = (char **) malloc ((result->anzahl + 1) * sizeof (char *))) == NULL) {
    free (result);
    return NULL;
  }
  for (index=0; index<result->anzahl; index++)
    if ((result->elements[index] = freads (datei, ptr)) == NULL)
      if (result->elements[index] = malloc (1 * sizeof (char)))
	(result->elements[index])[0] = '\0';

  result->elements[result->anzahl] = NULL;

  return result;
}

/****************************************************************************
***** Killer  -    Die folgenden Funktionen dienen dem Loeschen der     *****
*****              Resourcen (Speicherplatz freigeben)                  *****
*****************************************************************************/

void KillSboxResource (SBOX *tokill)
{
  int index;

  if (tokill!=NULL) {
    for (index=0; index<tokill->anzahl; index++)
      free (tokill->items[index]);

    free (tokill->title);
    free (tokill->HelpContext);

    free (tokill->items);
    free (tokill->resid);

    free (tokill);
  }
}

void KillMenuResource (MENU *tokill)
{
  int index;

  if (tokill != NULL) {
    for (index=0; index<tokill->anzahl; index++) {
      free (tokill->submenu[index]);
      free (tokill->text[index]);
    }
    free (tokill->HelpContext);
    free (tokill);
  }
}

void KillTextResource (TEXT *tokill)
{
  if (tokill != NULL) {
    free (tokill->title);
    free (tokill->HelpContext);

    if (tokill->offset != -1)
      free (tokill->filename);

    free (tokill);
  }
}

void KillMsgResource (MESSAGE *tokill)
{
  if (tokill != NULL) {
    free (tokill->text);
    free (tokill->HelpContext);
    free (tokill);
  }
}

void KillFormResource (FORM *tokill)
{
  int index;

  if (tokill != NULL) {
    for (index=0; index<tokill->anzahl; index++) {
      free (tokill->label[index]);
      free (tokill->stand[index]);
    }
    free (tokill->title);
    free (tokill->HelpContext);

    free (tokill->label);  free (tokill->stand);
    free (tokill->ftype);  free (tokill->resid);
    free (tokill->flen);   free (tokill->fxpos);
    free (tokill->fypos);  free (tokill->lxpos);
    free (tokill->lypos);

    free (tokill);
  }
}

void KillTvecResource (TVECTOR *tokill)
{
  int index;

  if (tokill != NULL) {
    for (index=0; index<tokill->anzahl; index++)
      free (tokill->elements[index]);
  
    free (tokill->elements);
    free (tokill);
  }
}

void KillResource (RESOURCE *tokill)
{
  if (tokill != NULL) {
    KillResource (tokill->LEFT);
    KillResource (tokill->RIGHT);

    switch (tokill->type) {
    case RES_SBOX:
      KillSboxResource (tokill->data.sbox);
      break;
    case RES_MENU:
      KillMenuResource (tokill->data.menu);
      break;
    case RES_TEXT:
      KillTextResource (tokill->data.text);
      break;
    case RES_MSG:
      KillMsgResource  (tokill->data.msg);
      break;
    case RES_FORM:
      KillFormResource (tokill->data.form);
      break;
    case RES_TVEC:
      KillTvecResource (tokill->data.tvec);
      break;
    }
    free (tokill->rname);
    free (tokill);
  }
}

/****************************************************************************
***** AddResource  Einfuegen einer Resource in den BTree                *****
*****************************************************************************/

RESOURCE *AddResource (RESOURCE *old, RESOURCE *new)
{
  if (old == NULL)
    return new;

  if (strcmp (old->rname, new->rname) < 0)
    old->LEFT  = AddResource (old->LEFT, new);
  else if (strcmp (old->rname, new->rname) > 0)
    old->RIGHT = AddResource (old->RIGHT, new);
  else if (old->type < new->type)
    old->LEFT  = AddResource (old->LEFT, new);
  else if (old->type > new->type)
    old->RIGHT = AddResource (old->RIGHT, new);
  else {
    new->LEFT  = old->LEFT;  old->LEFT  = NULL;
    new->RIGHT = old->RIGHT; old->RIGHT = NULL;

    KillResource (old);
    return new;
  }

  return old;
}

static RESOURCE *MergeTrees (RESOURCE *old, RESOURCE *new)
{
  RESOURCE *tmp;

  if (new == NULL)
    return old;

  tmp = MergeTrees (old, new->LEFT);  new->LEFT  = NULL;
  tmp = MergeTrees (tmp, new->RIGHT); new->RIGHT = NULL;
  
  return AddResource (tmp, new);
}

void AddMoreResources (RESOURCE *more)
{
  mainres = MergeTrees (mainres, more);
}

/****************************************************************************
***** Loader       Einlesen eines kompletten Resourcefiles              *****
*****              Der BTree wird aufgebaut                             *****
*****************************************************************************/

RESOURCE *LoadResource (char *filename, char *ptr)
{
  char c, *rtype, *HelpContext, *fname, *nptr;
  RESOURCE *result=NULL, *tmp=NULL;
  int fail, kill;
  FILE *datei;

  if (filename == NULL) {
    datei = NULL;

    if ((nptr = ptr) == NULL)
      return NULL;
  }
  else {
    if ((datei = fopen (filename, "rt")) == NULL)
      return NULL;
    nptr = NULL;
  }

  do {
    c = freadc (datei, &nptr);
    while (!FEOFC (datei, &nptr) && c != '[')
      c = freadc (datei, &nptr);
    if (FEOFC (datei, &nptr))
      break;

    if ((c = freadc (datei, &nptr)) == '^') {
      c = freadc (datei, &nptr);
      
      while (!FEOFC (datei, &nptr) && isalpha (c)) {
	for (fail=0, kill=1; accpref[fail]; fail++)
	  if (c == accpref[fail])
	    kill = 0;
	
	c = freadc (datei, &nptr);
      }
    }
    else {
      kill = 0;
      fungetc (c);
    }
    if (tmp == NULL) {
      if ((tmp = (RESOURCE *) malloc (sizeof (RESOURCE))) == NULL) {
	if (datei)
	  fclose (datei);
	return result;
      }
      tmp->LEFT  = NULL;
      tmp->RIGHT = NULL;
    }

    tmp->rname = freads (datei, &nptr);
    c          = freadc (datei, &nptr); /* Doppelpunkt lesen */
    rtype      = freads (datei, &nptr);
    c          = freadc (datei, &nptr); /* \] oder : lesen */

    if (c == ':') { /* Es folgt ein HelpContext */
      HelpContext = freads (datei, &nptr);
      c           = freadc (datei, &nptr); /* Nu aber \] */
    }
    else
      HelpContext = NULL;

    do {
      c = freadc (datei, &nptr); /* \n lesen */
    }
    while (c != '\n');

    fail       = 0;

    if (strcmp (rtype, "Resource") == 0) {
      fname  = freads (datei, &nptr);
      result = MergeTrees (result, LoadResource (fname, NULL));
      fail   = 1;
      free (fname);
    }
    else if (strcmp (rtype, "Text") == 0) {
      if ((tmp->data.text = LoadTextResource (datei, &nptr, filename, tmp->rname, HelpContext)) == NULL)
	fail = 1;
      else
	tmp->type = RES_TEXT;
    }
    else if (strcmp (rtype, "Select") == 0) {
      if ((tmp->data.sbox = LoadSboxResource (datei, &nptr, tmp->rname, HelpContext)) == NULL)
	fail = 1;
      else
	tmp->type = RES_SBOX;
    }
    else if (strcmp (rtype, "Menu") == 0) {
      if ((tmp->data.menu = LoadMenuResource (datei, &nptr, tmp->rname, HelpContext)) == NULL)
	fail = 1;
      else
	tmp->type = RES_MENU;
    }
    else if (strcmp (rtype, "Message") == 0) {
      if ((tmp->data.msg  = LoadMsgResource  (datei, &nptr, tmp->rname, HelpContext)) == NULL)
	fail = 1;
      else
	tmp->type = RES_MSG;
    }
    else if (strcmp (rtype, "Form") == 0) {
      if ((tmp->data.form = LoadFormResource (datei, &nptr, tmp->rname, HelpContext)) == NULL)
	fail = 1;
      else
	tmp->type = RES_FORM;
    }
    else if (strcmp (rtype, "Tvector") == 0) {
      if ((tmp->data.tvec = LoadTvecResource (datei, &nptr, tmp->rname, HelpContext)) == NULL)
	fail = 1;
      else
	tmp->type = RES_TVEC;
    }
    else if (strcmp (rtype, "Parameter") == 0) {
      LoadExternResource (datei, &nptr, tmp->rname, rtype);
      fail = 1;
    }
    else {
      LoadExternResource (datei, &nptr, tmp->rname, rtype);
      fail = 1;
    }

    if (fail) {
      if (HelpContext != NULL)
	free (HelpContext);
      if (tmp->rname != NULL)
	free (tmp->rname);
    }
    else {
      if (kill)
	KillResource (tmp);
      else
	result = AddResource (result, tmp);

      tmp = NULL;
    }
    
    free (rtype);
  }
  while (TRUE);

  if (datei)
    fclose (datei);

  return result;
}

/****************************************************************************
***** GetResource  Abfragen von Resourcen aus dem Binaeren Baum         *****
*****************************************************************************/

RESOURCE *GetResource (char *rname, int rtype)
{
  RESOURCE *res = mainres;
  int comp;

  if (rname == NULL)
    return NULL;

  while (res != NULL) {
    comp = strcmp (res->rname, rname);

    if (comp < 0)
      res = res->LEFT;
    else if (comp > 0)
      res = res->RIGHT;
    else if (res->type < rtype)
      res = res->LEFT;
    else if (res->type > rtype)
      res = res->RIGHT;
    else
      break;
  }

  if (res == NULL && strcmp(rname, "ResNotFound"))
    InfoUser (GetMsgResource ("ResNotFound"), rname);

  return res;
}

SBOX *GetSboxResource (char *rname)
{
  RESOURCE *get = GetResource (rname, RES_SBOX);
  
  if (get == NULL)
    return NULL;
  else
    return get->data.sbox;
}

MENU *GetMenuResource (char *rname)
{
  RESOURCE *get = GetResource (rname, RES_MENU);
  
  if (get == NULL)
    return NULL;
  else
    return get->data.menu;
}

TEXT *GetTextResource (char *rname)
{
  RESOURCE *get = GetResource (rname, RES_TEXT);

  if (get == NULL)
    return NULL;
  else
    return get->data.text;
}

FORM *GetFormResource (char *rname)
{
  RESOURCE *get = GetResource (rname, RES_FORM);
  
  if (get == NULL)
    return NULL;
  else
    return get->data.form;
}

char *GetMsgResource (char *rname)
{
  RESOURCE *get = GetResource (rname, RES_MSG);
  
  if (get == NULL)
    return NULL;
  else
    return get->data.msg->text;
}

char **GetTvecResource (char *rname)
{
  RESOURCE *get = GetResource (rname, RES_TVEC);

  if (get == NULL)
    return NULL;
  else
    return get->data.tvec->elements;
}

/****************************************************************************/
/*
 * Start der Bildschirmfunktionen
 * Die Resourcen uebernehmen den Bildschirm !
 *
 * Farbpaare werden gesetzt, die Resourcedatei wird gelesen
 **/

static void InitTerm (void)
{
#ifdef NEWTERM
  if (actterm)
    FinishTerm ();

  set_term (actterm = newterm (NULL, stdout, stdin));
#endif

#ifdef INITSCR
  initscr  ();
#endif
  noecho   ();
  cbreak   ();
  keypad   (stdscr, TRUE);

#ifndef _NOCOLORS_
  if (monoflag) {
    if (start_color () == OK) {
      init_pair (1, COLOR_WHITE, COLOR_BLACK);
      init_pair (2, COLOR_WHITE, COLOR_BLACK);
      init_pair (3, COLOR_WHITE, COLOR_BLACK);
      init_pair (4, COLOR_WHITE, COLOR_BLACK);
      init_pair (5, COLOR_WHITE, COLOR_BLACK);
    }
  }
  else {
    if (start_color () == OK) {
      init_pair (1, COLOR_GREEN, COLOR_BLACK);
      init_pair (2, COLOR_WHITE, COLOR_BLUE);
      init_pair (3, COLOR_BLACK, COLOR_BLUE);
      init_pair (4, COLOR_YELLOW,COLOR_BLUE);
      init_pair (5, COLOR_GREEN, COLOR_BLUE);
    }
  }
#endif
}

static void FinishTerm (void)
{
#ifdef NEWTERM
  if (actterm == NULL)
    return;
#endif

  wattrset     (stdscr, COLOR_PAIR(0));
  wfill        (stdscr);

  update_panels();
  doupdate     ();

#ifdef ENDWIN
  endwin       ();
#endif

#ifdef DELSCREEN
  delscreen (actterm);
#endif
  actterm = NULL;
}

int StartCurses (char *rfname, char *ptr, int mono)
{
  struct termio term;
  void __R_LoadIntern (void);

  monoflag = mono;

  InitTerm ();
/*
 * Ein wenig Systemquatsch; Normalerweise wird mit der <ENTF>-Taste
 * ein SIGINT ausgeloest anstelle ein Zeichen zu loeschen (<DEL>).
 * Das ganze wird abgeschaltet und auf Control-C umgelenkt.
 **/
  ioctl (fileno (stdin), TCGETA, &term);
  term.c_cc[VINTR] = CTRL('C');
  ioctl (fileno (stdin), TCSETA, &term);

/*
 * Installation eines Signalhandlers fuer SIGINT (Control-C)
 * Dort wird einfach isBreak auf True gesetzt.
 **/
  isBreak = 0; BlockBreak = 0;

#ifdef SIGACTION
  sigemptyset (&act.sa_mask);
  act.sa_handler = SigSetBreak;
  sigaction (SIGINT, &act, NULL);
#else
  signal (SIGINT, SigSetBreak);
#endif

  mainres = NULL;

  if (rfname)
    AddMoreResources (LoadResource (rfname, NULL));
  
  if (ptr)
    AddMoreResources (LoadResource (NULL, ptr));

  __R_LoadIntern ();

/*
 * Hilfetasten einbinden
 **/

  SetKeyHook ('^',      DisplayHelp);
  SetKeyHook (KEY_F(1), DisplayHelp);
  SetKeyAlias(KEY_HELP, KEY_F(1));

  return TRUE;
}

void EndCurses (void)
{
/*
 * Signalhandler deinstallieren
 **/
#ifdef SIGACTION
  act.sa_handler = SIG_DFL;
  sigaction (SIGINT, &act, NULL);
#else
  signal (SIGINT, SIG_DFL);
#endif

/*
 * Hilfetasten abklemmen
 **/
  DelKeyAlias(KEY_HELP);
  DelKeyHook ('^');
  DelKeyHook (KEY_F(1));

/*
 * Resourcen freigeben, Bildschirm loeschen
 **/
  FinishTerm   ();
  KillResource (mainres);

  mainres = NULL;
}

int DoResize (int lines, int cols)
{
  SCREEN *nt;
  char *s1, *s2;
  
  if ((s1 = malloc (80)) == NULL)
    return FALSE;
  if ((s2 = malloc (80)) == NULL) {
    free (s1);
    return FALSE;
  }

  FinishTerm ();

  sprintf (s1, "COLUMNS=%d", cols);  putenv (s1);
  sprintf (s2, "LINES=%d", lines); putenv (s2);

  InitTerm ();

  return TRUE;
}

int GetWinSize (int *lines, int *cols)
{
  FILE *datei;
  char string[80], *ptr;

  if ((datei = popen ("resize -u", "r")) == NULL)
    return 0;
  
  *lines = *cols = 0;

  while (!feof (datei) && (*lines == 0 || *cols == 0)) {
    if (fgets (string, 80, datei) == NULL)
      break;
    if ((ptr = strchr (string, '=')) == NULL)
      continue;
    if (strncmp (string, "LINES", 5) == 0)
      *lines = atoi (ptr+1);
    else if (strncmp (string, "COLUMNS", 7) == 0)
      *cols  = atoi (ptr+1);
  }
  pclose (datei);

  if (*lines == 0 || *cols == 0)
    return 0;

  return 1;
}

int HavetoResize (void)
{
  static time_t lastpoll = 0;
  static int fail = 0;
  int ncol, nlines;

  if (fail)
    return 0;
  if (time(NULL)-lastpoll < 10)
    return 0;
  if (!GetWinSize (&nlines, &ncol)) {
    fail = 1;
    return 0;
  }
  lastpoll = time(NULL);

  if (COLS != ncol || LINES != nlines)
    return 1;

  return 0;
}

/****************************************************************************/
/*
 * Lesen eines Zeichens von der Tastatur
 *
 * Wird eine Taste gefunden, die per 'SetKeyHook' an eine Funktion gebunden
 * ist, so wird die Funktion aufgerufen und die naechste Taste gelesen
 * Wird eine Taste gefunden, die mit 'SetKeyAlias' an eine andere Taste ge-
 * fesselt ist, so wird der Code ersetzt
 * Nicht zuletzt werden auch Macros ersetzt.
 **/

int InputChar (void)
{
  static char *yc = "BOOB", *zc = "HFMJ", *xc = "GQQ";
  static int yi=0, zi=0, xi=0;
  int result, index, *iter;

 newchar:
/*
 * Wenn Control-C aufgetreten ist, dann ist es hier sowieso
 * zu spaet -> ignorieren
 **/
  if (HaveABreak())
    BreakProcessed ();

  update_panels ();
  doupdate      ();

 reget:
  if (KB_Count > 0) {
    result = KeyBuffer[0]; KB_Count--;
    memcpy (KeyBuffer, KeyBuffer+1, KB_Count * sizeof (int));
  }
  else {
    BlockBreak = 1;
    result = wgetch(stdscr);

#ifdef SLOWINPUT
    if (result == '\x1b') {
      nodelay (stdscr, TRUE);
      sleep   (1);
      if (wgetch (stdscr) == '[') {
	switch (wgetch (stdscr)) {
	case 'A': result = KEY_UP;    break;
	case 'B': result = KEY_DOWN;  break;
	case 'C': result = KEY_RIGHT; break;
	case 'D': result = KEY_LEFT;  break;
	case 'F': result = KEY_END;   break;
	case 'G': result = KEY_NPAGE; break;
	case 'H': result = KEY_BEG;   break;
	case 'I': result = KEY_PPAGE; break;
	case 'L': result = KEY_IC;    break;
	case 'Z': result = KEY_BTAB;  break;
	default:  result = '\x1b';    break;
	}
      }
      nodelay (stdscr, FALSE);
    }
#endif /* SLOWINPUT */

    BlockBreak = 0;
  }

  if (result <= 0)
    goto newchar;

 retest:
/*
 * Was das soll? Wird nicht verraten! BITTE NICHT AENDERN !!!
 **/

  if (result==CTRL(*(xc+xi)-1)){xi++;if(!*(xc+xi)){xi=0;itsanfpp();goto reget;}else if(xi>1)goto reget;}else xi=0;
  if (result==CTRL(*(yc+yi)-1)){yi++;if(!*(yc+yi)){yi=0;itsanfpp();goto reget;}else if(yi>1)goto reget;}else yi=0;
  if (result==CTRL(*(zc+zi)-1)){zi++;if(!*(zc+zi)){zi=0;itsanfpp();goto reget;}else if(zi>1)goto reget;}else zi=0;

  for (index=KM_Count-1; index>=0; index--)
    if (result == *KeyMacros[index]) {
      iter = KeyMacros[index] + 1;

      while (*iter)
	KeyBuffer[KB_Count++] = *iter++;
      goto reget;
    }

  for (index=KA_Count-1; index>=0; index--)
    if (result == KeyAlias[index][0]) {
      result = KeyAlias[index][1];
      goto retest;
    }

  for (index=KH_Count-1; index>=0; index--)
    if (result == KeyHooks[index]) {
      (*KH_Funcs[index]) ();
      goto newchar;
    }

  return result;
}

/*
 * Fuelle einen gesamtes Fenster mit Leerzeichen
 * (d.h. mit der aktuellen Hintergrundfarbe)
 **/

void wfill (WINDOW *win)
{
  int yres, xres, count;

  getmaxyx (win, yres, xres);
  wmove    (win,    0,    0);

  for (; yres; yres--)
    for (count=xres; count; count--)
      waddch (win, ' ');
}

/*
 * Dasselbe, aber mit einem Teilbereich des Fensters (inklusive)
 **/

void wfillrgn (WINDOW *win, int y1, int x1, int y2, int x2)
{
  int xc, yc;

  for (yc=y1; yc<=y2; yc++) {
    wmove (win, yc, x1);
    for (xc=x1; xc<=x2; xc++)
      waddch (win, ' ');
  }
}

/*
 * Ziehe eine horizontale Linie in der Hoehe ywo ueber das gesamte Fenster
 **/

void whorline (WINDOW *win, int ywo)
{
  int yres, xres;
  int count;

  getmaxyx (win, yres, xres);
  wmove    (win, ywo,  0);

  for (count=1; count<xres; count++)
    waddch (win, ACS_HLINE);

  mvwaddch (win, ywo, 0,      ACS_SSSB);
  mvwaddch (win, ywo, xres-1, ACS_SBSS);
}

/*
 * Zeichne ein beliebiges Kaestchen im Fenster
 **/

void wbox (WINDOW *win, int y1, int x1, int y2, int x2)
{
  int ycount, xcount;

  if (x1 == x2 && y1 == y2)
    return;

  if (x1 == x2)
    mvwaddch (win, y1, x1, ACS_SBSB);
  else if (y1 == y2)
    mvwaddch (win, y1, x1, ACS_BSBS);
  else
    mvwaddch (win, y1, x1, ACS_BSSB);
  
  for (xcount=x1+1; xcount<x2; xcount++)
    waddch (win, ACS_HLINE);
  
  if (x1 == x2);
  else if (y1 == y2) {
    waddch (win, ACS_BSBS);
    return;
  }
  else
    waddch (win, ACS_BBSS);

  for (ycount=y1+1; ycount<y2; ycount++) {
    mvwaddch (win, ycount, x1, ACS_VLINE);
    mvwaddch (win, ycount, x2, ACS_VLINE);
  }
  if (x1 == x2)
    mvwaddch (win, y2, x1, ACS_SBSB);
  else
    mvwaddch (win, y2, x1, ACS_SSBB);

  for (xcount=x1+1; xcount<x2; xcount++)
    waddch (win, ACS_HLINE);

  if (x1 != x2)
    waddch (win, ACS_SBBS);
}

/*
 * Umwandeln von Linienzeichen des ACS in eine Zahl
 * erg & 1 == TRUE: Zeichen besitzt Abzweigung nach oben
 * erg & 2 == TRUE:    ""      ""       ""      ""  rechts
 * erg & 4 == TRUE:    ""      ""       ""      ""  unten
 * erg & 8 == TRUE:    ""      ""       ""      ""  links
 **/

int acs2int (chtype src)
{
  int res = 0;

  if (src == ACS_BBSS)
    res = AI_DOWN | AI_LEFT; 
  else if (src == ACS_BSBS)
    res = AI_RIGHT | AI_LEFT;
  else if (src == ACS_BSSB)
    res = AI_RIGHT | AI_DOWN;
  else if (src == ACS_SBBS)
    res = AI_UP | AI_LEFT;
  else if (src == ACS_SBSB)
    res = AI_UP | AI_DOWN;
  else if (src == ACS_SSBB)
    res = AI_UP | AI_RIGHT;
  else if (src == ACS_SSSB)
    res = AI_UP | AI_RIGHT | AI_DOWN;
  else if (src == ACS_SSBS)
    res = AI_UP | AI_RIGHT | AI_LEFT;
  else if (src == ACS_SBSS)
    res = AI_UP | AI_DOWN | AI_LEFT;
  else if (src == ACS_BSSS)
    res = AI_RIGHT | AI_DOWN | AI_LEFT;
  else if (src == ACS_SSSS)
    res = AI_UP | AI_RIGHT | AI_DOWN | AI_LEFT;

  return res;
}

/*
 * Das umgekehrte, wandelt eine auf diese Weise gebildete Zahl
 * in das entsprechende Linienzeichen um
 **/

chtype int2acs (int src)
{
  chtype res = (chtype) ' ';

  if ((src & (AI_UP | AI_DOWN | AI_RIGHT | AI_LEFT)) == (AI_UP | AI_DOWN | AI_RIGHT | AI_LEFT))
    res = ACS_SSSS;
  else if ((src & (AI_RIGHT | AI_DOWN | AI_LEFT)) == (AI_RIGHT | AI_DOWN | AI_LEFT))
    res = ACS_BSSS;
  else if ((src & (AI_UP | AI_DOWN | AI_LEFT)) == (AI_UP | AI_DOWN | AI_LEFT))
    res = ACS_SBSS;
  else if ((src & (AI_RIGHT | AI_UP | AI_LEFT)) == (AI_RIGHT | AI_UP | AI_LEFT))
    res = ACS_SSBS;
  else if ((src & (AI_RIGHT | AI_UP | AI_DOWN)) == (AI_RIGHT | AI_UP | AI_DOWN))
    res = ACS_SSSB;
  else if ((src & (AI_RIGHT | AI_UP)) == (AI_RIGHT | AI_UP))
    res = ACS_SSBB;
  else if ((src & (AI_UP | AI_DOWN)) == (AI_UP | AI_DOWN))
    res = ACS_SBSB;
  else if ((src & (AI_UP | AI_LEFT)) == (AI_UP | AI_LEFT))
    res = ACS_SBBS;
  else if ((src & (AI_RIGHT | AI_DOWN)) == (AI_RIGHT | AI_DOWN))
    res = ACS_BSSB;
  else if ((src & (AI_RIGHT | AI_LEFT)) == (AI_RIGHT | AI_LEFT))
    res = ACS_BSBS;
  else if ((src & (AI_DOWN | AI_LEFT)) == (AI_DOWN | AI_LEFT))
    res = ACS_BBSS;

  return res;
}

/*
 * Wandelt ein Grafikzeichen in ein normales Textzeichen um
 * (so sinnvoll wie moeglich)
 **/

int simplechar (int zeichen)
{
  int normal;

  zeichen &= (A_CHARTEXT | A_ALTCHARSET);

  if ((normal = acs2int (zeichen)) == 0)
    return zeichen;
  else if (normal == (AI_UP | AI_DOWN))
    return '|';
  else if (normal == (AI_LEFT | AI_RIGHT))
    return '-';
  else
    return '+';
}

/*
 * WinLineAddChar: Eine auf gleiche Weise wie oben erzeugte Zahl
 * wird zum Zeichen auf genannter Position hinzugefuegt, also z.B. 
 * eine weitere Abzweigung
 **/

void wladdch (WINDOW *win, int y1, int x1, int type)
{
  int old = acs2int (mvwinch (win, y1, x1) & 
		     (A_CHARTEXT | A_ALTCHARSET));
  int xres, yres, new=0;

  getmaxyx (win, yres, xres);

  if (old == (AI_LEFT  | AI_RIGHT | AI_UP  ) ||
      old == (AI_LEFT  | AI_RIGHT | AI_DOWN) ||
      old == (AI_LEFT  | AI_UP    | AI_DOWN) ||
      old == (AI_RIGHT | AI_UP    | AI_DOWN) ||
      old == (AI_RIGHT | AI_UP    | AI_DOWN | AI_LEFT))
    new = old;
  else {
    if (old & AI_LEFT) {
      if (x1>0 && acs2int (mvwinch (win, y1, x1-1) &
			   (A_CHARTEXT | A_ALTCHARSET)) & AI_RIGHT)
	new |= AI_LEFT;
    }
    if (old & AI_UP) {
      if (y1>0 && acs2int (mvwinch (win, y1-1, x1) &
			   (A_CHARTEXT | A_ALTCHARSET)) & AI_DOWN)
	new |= AI_UP;
    }
    if (old & AI_RIGHT) {
      if (x1+1<xres && acs2int (mvwinch (win, y1, x1+1) &
				(A_CHARTEXT | A_ALTCHARSET)) & AI_LEFT)
	new |= AI_RIGHT;
    }
    if (old & AI_DOWN) {
      if (y1+1<yres && acs2int (mvwinch (win, y1+1, x1) &
				(A_CHARTEXT | A_ALTCHARSET)) & AI_UP)
	new |= AI_DOWN;
    }
  }
  new |= type;
  mvwaddch (win, y1, x1, int2acs (new));
}

/*
 * Zeichnen einer beliebigen Linie oder Box auf den Bildschirm
 * Das Besondere: Ueberschneidungen der Linien werden beruecksichtigt,
 * d.h. wenn eine horizontale Linie eine vertikale schneidet, entsteht
 * ein komplettes Kreuz
 **/

void wline (WINDOW *win, int y1, int x1, int y2, int x2)
{
  int temp, index;
  int ymax, xmax;
  int old, new;

  getmaxyx (win, ymax, xmax);

  if (y1 == y2) {
    if (x1 > x2) {
      temp = x1; x1 = x2; x2 = temp; }

    index = x1;

    while (index <= x2) {
      new = old = acs2int (mvwinch (win, y1, index) & 
			   (A_CHARTEXT | A_ALTCHARSET));

      if (old == (AI_LEFT | AI_DOWN)) {
	if (index != x2)
	  new |= AI_RIGHT;
      } 
      else if (old == (AI_DOWN | AI_RIGHT)) {
	if (index != x1)
	  new |= AI_LEFT;
      } 
      else if (old == (AI_UP | AI_LEFT)) {
	if (index != x2)
	  new |= AI_RIGHT;
      } 
      else if (old == (AI_UP | AI_DOWN)) {
	if (index == x1) {
	  new = AI_RIGHT;
	  if (y1>0 && acs2int (mvwinch (win, y1-1, index) & 
			       (A_CHARTEXT | A_ALTCHARSET)) & AI_DOWN)
	    new |= AI_UP;
	  if (y1+1<ymax && acs2int (mvwinch (win, y1+1, index) &
				    (A_CHARTEXT | A_ALTCHARSET)) & AI_UP)
	    new |= AI_DOWN;
	  if (new == AI_RIGHT)
	    new |= AI_UP | AI_DOWN;
	}
	else if (index == x2) {
	  new = AI_LEFT;
	  if (y1>0 && acs2int (mvwinch (win, y1-1, index) & 
			       (A_CHARTEXT | A_ALTCHARSET)) & AI_DOWN)
	    new |= AI_UP;
	  if (y1+1<ymax && acs2int (mvwinch (win, y1+1, index) &
				    (A_CHARTEXT | A_ALTCHARSET)) & AI_UP)
	    new |= AI_DOWN;
	  if (new == AI_LEFT)
	    new |= AI_UP | AI_DOWN;
	}
	else {
	  new = AI_LEFT | AI_RIGHT;
	  if (y1>0 && acs2int (mvwinch (win, y1-1, index) & 
			       (A_CHARTEXT | A_ALTCHARSET)) & AI_DOWN)
	    new |= AI_UP;
	  if (y1+1<ymax && acs2int (mvwinch (win, y1+1, index) &
				    (A_CHARTEXT | A_ALTCHARSET)) & AI_UP)
	    new |= AI_DOWN;
	}
      }
      else if (old == (AI_UP | AI_RIGHT)) {
	if (index != x1)
	  new |= AI_LEFT;
      }
      else if (old == (AI_UP | AI_RIGHT | AI_DOWN)) {
	if (index != x1)
	  new |= AI_LEFT;
      } 
      else if (old == (AI_UP | AI_DOWN | AI_LEFT)) {
	if (index != x2)
	  new |= AI_RIGHT;
      }
      else if (old & (AI_LEFT | AI_RIGHT) == (AI_LEFT | AI_RIGHT))
	new = old;
      else
	new = AI_LEFT | AI_RIGHT;

      mvwaddch (win, y1, index++, int2acs (new));
    }
  }
  else if (x1 == x2) {
    if (y1 > y2) {
      temp = y1; y1 = y2; y2 = temp; }

    index = y1;

    while (index <= y2) {
      new = old = acs2int (mvwinch (win, index, x1) & 
			   (A_CHARTEXT | A_ALTCHARSET));

      if (old == (AI_DOWN | AI_LEFT)) {
	if (index != y1)
	  new |= AI_UP;
      }
      else if (old == (AI_LEFT | AI_RIGHT)) {
	if (index == y1) {
	  new = AI_DOWN;
	  if (x1>0 && acs2int (mvwinch (win, index, x1-1) &
			       (A_CHARTEXT | A_ALTCHARSET)) & AI_RIGHT)
	    new |= AI_LEFT;
	  if (x1+1<xmax && acs2int (mvwinch (win, index, x1+1) &
				    (A_CHARTEXT | A_ALTCHARSET)) & AI_LEFT)
	    new |= AI_RIGHT;

	  if (new == AI_DOWN)
	    new |= AI_LEFT | AI_RIGHT;
	}
	else if (index == y2) {
	  new = AI_UP;
	  if (x1>0 && acs2int (mvwinch (win, index, x1-1) &
			       (A_CHARTEXT | A_ALTCHARSET)) & AI_RIGHT)
	    new |= AI_LEFT;
	  if (x1+1<xmax && acs2int (mvwinch (win, index, x1+1) &
				    (A_CHARTEXT | A_ALTCHARSET)) & AI_LEFT)
	    new |= AI_RIGHT;

	  if (new == AI_UP)
	    new |= AI_LEFT | AI_RIGHT;
	}
	else {
	  new = AI_UP | AI_DOWN;
	  if (x1>0 && acs2int (mvwinch (win, index, x1-1) &
			       (A_CHARTEXT | A_ALTCHARSET)) & AI_RIGHT)
	    new |= AI_LEFT;
	  if (x1+1<xmax && acs2int (mvwinch (win, index, x1+1) &
				    (A_CHARTEXT | A_ALTCHARSET)) & AI_LEFT)
	    new |= AI_RIGHT;
	}
      }
      else if (old == (AI_RIGHT | AI_DOWN)) {
	if (index != y1)
	  new |= AI_UP;
      }
      else if (old == (AI_UP | AI_LEFT)) {
	if (index != y2)
	  new |= AI_DOWN;
      }
      else if (old == (AI_UP | AI_RIGHT)) {
	if (index != y2)
	  new |= AI_DOWN;
      }
      else if (old == (AI_UP | AI_RIGHT | AI_LEFT)) {
	if (index != y2)
	  new |= AI_DOWN;
      }
      else if (old == (AI_RIGHT | AI_DOWN | AI_LEFT)) {
	if (index != y1)
	  new |= AI_UP;
      }
      else if (old & (AI_UP | AI_DOWN) == (AI_UP | AI_DOWN))
	new = old;
      else
	new = AI_UP | AI_DOWN;
      
      mvwaddch (win, index++, x1, int2acs (new));
    }
  }
  else {
    wline (win, y1, x1, y1, x2);
    wline (win, y2, x1, y2, x2);
    wline (win, y1, x1, y2, x1);
    wline (win, y1, x2, y2, x2);
  }
}

/*
 * Ausgeben einer Textzeile in Hoehe ypos
 * links- rechts- buendig, zentriert, blocksatz
 **/

void woutline (WINDOW *win, char *text, int ypos, int type)
{
  int length, yres, xres, count, spaces=0, idx;
  char *tmp = text;

  if (text == NULL)
    return;

  getmaxyx (win, yres, xres);
  length = strlen (text);

  switch (type) {
  case AL_LEFT:
    mvwaddstr (win, ypos, 0, text);
    break;
  case AL_RIGHT:
    mvwaddstr (win, ypos, xres-length, text);
    break;
  case AL_CENTER:
    mvwaddstr (win, ypos, (xres-length)/2, text);
    break;
  case AL_BLOCK:
    wmove (win, ypos, 0);
    count = xres - length;
    while (*tmp)
      if (*tmp++ == ' ')
	spaces++;
    while (*text) {
      waddch (win, *text);
      if (*text == ' ') {
	for (idx = 0; idx < count / spaces; idx++)
	  waddch (win, ' ');
	count -= count / spaces;
	spaces--;
      }
      text++;
    }
    break;
  }
}

/*
 * Ausgabe eines Strings in einem Fenster an einer bestimmten Position
 * xpos darf dabei auch negativ sein, weiterhin wird der Text am rechten
 * Ende des Fensters abgeschnitten
 **/

void wprintstr (WINDOW *win, char *string, int ypos, int xpos)
{
  int length, yres, xres, count;
  char *ptr = string;

  getmaxyx (win, yres, xres);
  length = strlen (string);

  if (ypos < 0 || ypos >= yres)
    return;
  if (xpos + length < 0 || xpos >= xres)
    return;

  if (xpos < 0 && xpos + length < xres)
    mvwaddstr (win, ypos, 0, string - xpos);
  else if (xpos > 0 && xpos + length < xres)
    mvwaddstr (win, ypos, xpos, string);
  else {
    if (xpos < 0) {
      ptr   = string - xpos;
      count = 0;
    }
    else
      count = xpos;

    wmove (win, ypos, count);
    while (*ptr && count < xres) {
      waddch (win, *ptr); count++; ptr++;
    }
  }
}

/*
 * Wiederherstellung eines verpfuschten Bildschirmes, per KeyHook an
 * CTRL-? gehaengt
 **/

void RedrawScreen (void)
{
  wclear        (curscr);
  update_panels ();
  doupdate      ();
}

/****************************************************************************/
/* 
 * Anzeigen eines Hilfetextes
 * ist an KEY_F(1) ... gebunden
 **/

void DisplayHelp (void)
{
  if (HelpCount == -1 || HelpStack[HelpCount] == NULL)
    InfoUser (GetMsgResource ("Sorry"), GetMsgResource ("NoHelp"));
  else
    DisplayText (HelpStack[HelpCount]);
}

/*
 * Hilfeindex fuer Kontextsensitive Hilfe setzen
 **/

void SetHelpContext (char *Context)
{
  HelpStack[++HelpCount] = Context;
}

/*
 * Aktuelle Hilfeseiten verwerfen
 **/

void DelHelpContext (void)
{
  HelpCount--;
}

/*
 * Einbinden oder Ausbinden eines Tastaturmakros
 **/

void SetKeyMacro (int readkey, int *writekey)
{
  int count = 1, *iter = writekey;

  while (*iter++)
    count++;

  if ((iter = KeyMacros[KM_Count] = (int *) malloc ((count+1) * sizeof (int))) == NULL)
    return;
  
  *iter++ = readkey;

  do {
    *iter = *writekey++;
  }
  while (*iter++);

  KM_Count++;
}

void DelKeyMacro (int readkey)
{
  int index=KM_Count-1;

  while (index >= 0)
    if (*KeyMacros[index] == readkey)
      break;
    else
      index--;
  
  if (index < 0)
    return;

  free (KeyMacros[index]);
  KM_Count--;

  while (index < KM_Count) {
    KeyMacros[index] = KeyMacros[index+1];
    KeyMacros[index] = KeyMacros[index+1];
    index++;
  }
}

/*
 * Einbinden oder ausbinden eines Key-Alias
 **/

void SetKeyAlias (int readkey, int writekey)
{
  KeyAlias[KA_Count][0] = readkey;
  KeyAlias[KA_Count][1] = writekey;
  KA_Count++;
}

void DelKeyAlias (int readkey)
{
  int index=KA_Count-1;

  while (index >= 0)
    if (KeyAlias[index][0] == readkey)
      break;
    else
      index--;
  
  if (index < 0)
    return;

  KA_Count--;

  while (index < KA_Count) {
    KeyAlias[index][0] = KeyAlias[index+1][0];
    KeyAlias[index][1] = KeyAlias[index+1][1];
    index++;
  }
}

/*
 * Einbinden bzw. ausbinden einer Funktion an eine Taste
 **/

void SetKeyHook (int keycode, void (*function) (void))
{
  KeyHooks[KH_Count] = keycode;
  KH_Funcs[KH_Count] = function;
  KH_Count++;
}

void DelKeyHook (int keycode)
{
  int index=KH_Count-1;

  while (index >= 0)
    if (KeyHooks[index] == keycode)
      break;
    else
      index--;
  
  if (index < 0)
    return;

  KH_Count--;

  while (index < KH_Count) {
    KeyHooks[index] = KeyHooks[index+1];
    KH_Funcs[index] = KH_Funcs[index+1];
    index++;
  }
}

/*
 * Oeffnet eine 'Messagebox' mit Titelzeile
 * und stellt eine Meldung dar (wird selbst umgebrochen)
 **/

void InfoUser (char *title, char *meldung)
{
  char *temp = meldung, *t2;
  WINDOW *win, *sub;
  PANEL *pan;
  int y=2, c;

  if ((win = newwin (15, 52, (LINES-15)/2, (COLS-52)/2)) == NULL)
    return;
  if ((pan = new_panel (win)) == NULL) {
    delwin (win);
    return;
  }
  wattrset (win, COLOR_PAIR (3));
  wfill    (win);

  if ((sub = subwin (win, 9, 35, 3+(LINES-15)/2, 12+(COLS-52)/2)) == NULL) {
    del_panel (pan);
    delwin    (win);
    return;
  }

  box       (win, 0, 0);
  whorline  (win, 2);
  wattrset  (win, COLOR_PAIR (4));
  woutline  (win, title, 1, AL_CENTER);
  mvwaddstr (win,  4, 2, "        ");
  mvwaddstr (win,  5, 2, "   ##   ");
  mvwaddstr (win,  6, 2, "        ");
  mvwaddstr (win,  7, 2, "  ###   ");
  mvwaddstr (win,  8, 2, "   ##   ");
  mvwaddstr (win,  9, 2, "   ##   ");
  mvwaddstr (win, 10, 2, " ###### ");
  mvwaddstr (win, 11, 2, "        ");
  wattrset  (sub, COLOR_PAIR (2));

  do {
    while (*temp == ' ')
      temp++;
    t2    = temp;
    for (c=0; c<35; c++, temp++)
      if (*temp == '\0')
	break;
    if (*temp == '\0') {
      woutline (sub, t2, y, AL_LEFT);
      break;
    }
    while (*temp != ' ')
      temp--;
    *temp = '\0';
    woutline (sub, t2, y++, AL_LEFT);
    *temp = ' ';
  }
  while (TRUE);

  wattrset (win, COLOR_PAIR (4));
  woutline (win, GetMsgResource ("PressKey"), 13, AL_CENTER);

  SetHelpContext (NULL);

  do {
    c = InputChar ();
    if (c == '\n' || c == '\x1b' || c == KEY_F(0) || c == CTRL('X'))
      break;
  }
  while (TRUE);

  DelHelpContext ();

  del_panel     (pan);
  delwin        (sub);
  delwin        (win);
}

/****************************************************************************/

void InitMenu (char *rname)
{
  MENU *menue = GetMenuResource (rname);
  int index;

  if (menue == NULL)
    return;

  wmove    (stdscr, 1, 0);
  wattrset (stdscr, COLOR_PAIR (1));

  for (index=0; index<COLS; index++)
    waddch (stdscr, ACS_HLINE);

  wmove    (stdscr, 0, 0);
  wattrset (stdscr, COLOR_PAIR (2));

  for (index=0; index<COLS; index++)
    waddch (stdscr, ' ');

  for (index=0; index<menue->anzahl; index++)
    mvwaddstr (stdscr, menue->ypos[index], menue->xpos[index], menue->text[index]);

  wattrset (stdscr, COLOR_PAIR (0));
}

int DoSboxMain (SBOX *auswahl, int ypos, int xpos, int flag)
{
  int item, c, yflag;
  extern WINDOW *newwin (int, int, int, int);
  WINDOW *win;
  PANEL  *pan;

  if ((flag & SB_TITLED) && auswahl->title && auswahl->title[0])
    yflag = 4;
  else
    yflag = 2;

  switch (-ypos) {
  case AL_LEFT:  ypos=1; break;
  case AL_RIGHT: ypos= LINES-auswahl->ysize-yflag; break;
  case AL_CENTER:ypos=(LINES-auswahl->ysize-yflag) / 2; break;
  default:       break;
  }
  switch (-xpos) {
  case AL_LEFT:  xpos=1; break;
  case AL_RIGHT: xpos= COLS-auswahl->xsize-2; break;
  case AL_CENTER:xpos=(COLS-auswahl->xsize-2) / 2; break;
  default:       break;
  }

  if (auswahl != NULL) {
    if (ypos+auswahl->ysize+yflag > LINES)
      ypos = LINES-auswahl->ysize-yflag;
    if (xpos+auswahl->xsize > COLS)
      xpos = COLS-auswahl->xsize-2;

    if ((win = newwin (auswahl->ysize+yflag, auswahl->xsize+4, ypos, xpos)) == NULL)
      return -1;
    if ((pan = new_panel (win)) == NULL) {
      delwin (win);
      return 0;
    }
    wattrset (win, COLOR_PAIR (1));
    box      (win, 0, 0);

    if ((flag & SB_TITLED) && auswahl->title && auswahl->title[0]) {
      whorline  (win, 2);
      wattrset  (win, COLOR_PAIR (2));
      mvwaddstr (win, 1, 2, auswahl->title);
      yflag   = 3;
    }
    else
      yflag   = 1;

    wattrset (win, COLOR_PAIR (0));

    for (item=0; item < auswahl->anzahl; item++)
      mvwaddstr (win, item+yflag, 2, auswahl->items[item]);
  }

  item = 0;

  if (auswahl != NULL && auswahl->HelpContext != NULL)
    SetHelpContext (auswahl->HelpContext);

  do {
    if (auswahl != NULL) {
      wattrset (win, A_REVERSE | A_BOLD | COLOR_PAIR (1));
      mvwaddch (win, item+yflag, 1, ' ');
      waddstr  (win, auswahl->items[item]);
      waddch   (win, ' ');
      wattrset (win, COLOR_PAIR (0));
    }

    do {
      c = InputChar ();
      if ((c == KEY_RIGHT || c == KEY_LEFT ||
	   c == 'l'       || c == 'h') && flag & SB_MAYMOVE)
	break;
      if (c == KEY_UP    || c == KEY_DOWN || c == 0x1b || 
	  c == '\n'      || c == KEY_F(0) || c == 'k'  ||
	  c == 'j'       || c == KEY_HOME || c == KEY_END ||
	  c == CTRL('X'))
	break;
    }
    while (TRUE);

    if (auswahl != NULL) {
      mvwaddch (win, item+yflag, 1, ' ');
      waddstr  (win, auswahl->items[item]);
      waddch   (win, ' ');
    }

    switch (c) {
    case 'j':
    case KEY_DOWN:
      if (auswahl != NULL) {
	if (item+1 < auswahl->anzahl)
	  item++;
	else
	  item = 0;
      }
      break;
    case KEY_END:
      if (auswahl != NULL)
	item = auswahl->anzahl - 1;
      break;
    case 'k':
    case KEY_UP:
      if (auswahl != NULL) {
	if (auswahl->ysize > 0 && item > 0)
	  item--;
	else
	  item = auswahl->anzahl - 1;
      }
      break;
    case KEY_HOME:
      if (auswahl != NULL)
	item = 0;
      break;
    }
    if (c == KEY_LEFT || c == KEY_RIGHT || c == 0x1b || c == '\n' ||
	c == KEY_F(0) || c == CTRL('X') || c == 'h'       || c == 'l')
      break;
  }
  while (TRUE);

  if (auswahl != NULL) {
    del_panel (pan);
    delwin    (win);
  }

  if (auswahl != NULL && auswahl->HelpContext != NULL)
    DelHelpContext ();

  switch (c) {
  case 'h':
  case KEY_LEFT:
    return SB_LEFT;
  case 'l':
  case KEY_RIGHT:
    return SB_RIGHT;
  case '\n':
    if (auswahl != NULL)
      return auswahl->resid[item];
    else
      return SB_NONE;
  case CTRL('X'):
  case '\x1b':
  case KEY_F(0):
    return SB_ESCAPE;
  }
  return -1; /* shouldn't be reached */
}

int DoSbox (char *rname)
{
  SBOX *select = GetSboxResource (rname);
  int xpos, ypos;
  int result;

  if (select == NULL)
    return -1;
  else {
    SetHelpContext (select->HelpContext);
    result = DoSboxMain (select, select->ypos, select->xpos, SB_TITLED);
    DelHelpContext ();
  }
  return result;
}

int DynSelectBox (int anzahl,
		  char *title, char *helpcontext,
		  char *ppage, char *npage,
		  char **items)
{
  int count, res, flag, top, result=-1, width, pnflag;
  int npm[3], ppm[3];
  SBOX *auswahl;

  if ((auswahl = (SBOX *) malloc (sizeof (SBOX))) == NULL) {
    InfoUser (GetMsgResource ("Sorry"), GetMsgResource ("ResFehler"));
    return 0;
  }
  memset (auswahl, '\0', sizeof (SBOX));

  auswahl->title       = title;
  auswahl->HelpContext = helpcontext;
  auswahl->ypos        = -3;
  auswahl->xpos        = -3;
  auswahl->rname       = NULL;

  npm[0] = KEY_END;  npm[1] = '\n'; npm[2] = '\0';
  ppm[0] = KEY_HOME; ppm[1] = '\n'; ppm[2] = '\0';

  SetKeyMacro (KEY_NPAGE, npm);
  SetKeyMacro (KEY_PPAGE, ppm);

  if ((auswahl->resid = (int *) malloc ((LINES-10)*sizeof(int))) == NULL)
    goto SASpanic;
  if ((auswahl->items = (char **) malloc ((LINES-10)*sizeof(char *))) == NULL)
    goto SASpanic;
  
  top = 0;

 SASNochMal:
  memset (auswahl->resid, 0, (LINES-10) * sizeof (int));
  memset (auswahl->items, 0, (LINES-10) * sizeof (char *));

  count   = 0;
  flag    = 0;
  pnflag  = 0;
  width   = strlen(auswahl->title);

  if (top > 0 && ppage) {
    auswahl->resid[count] = 998;
    auswahl->items[count] = ppage;

    if (strlen (auswahl->items[count]) > width)
      width = strlen(auswahl->items[count]);
    
    flag    = 1;
    pnflag |= 1;
    count++;
  }
  for (; count<(LINES-11) && top+count-flag<anzahl && items[top+count-flag]; count++) {
    auswahl->items[count] = items[top+count-flag];
    auswahl->resid[count] = 1000+count-flag;

    if (strlen (items[top+count-flag]) > width)
      width = strlen (items[top+count-flag]);
  }
  if (top+count-flag<anzahl && items[top+count-flag+1] && npage) {
    auswahl->resid[count] = 999;
    auswahl->items[count] = npage;

    if (strlen (auswahl->items[count]) > width)
      width = strlen (auswahl->items[count]);

    pnflag |= 2;
    count++;
  }
  auswahl->ysize   = count;
  auswahl->xsize   = width+2;
  auswahl->anzahl  = count;

  SetHelpContext   (auswahl->HelpContext);
  res = DoSboxMain (auswahl, auswahl->ypos, auswahl->xpos, SB_TITLED);
  DelHelpContext   ();

  if (res == IT_ESCAPE)
    result = SB_ESCAPE;
  else if (res == 998) {
    if (top > LINES-12)
      top -= LINES-12;
    else
      top  = 0;
    goto SASNochMal;
  }
  else if (res == 999) {
    if (top+LINES-12<anzahl && items[top+LINES-12])
      top += LINES-12-flag;
    else
      top += LINES-12-flag;
    goto SASNochMal;
  }
  else
    result = top+res;
  
 SASpanic:
  DelKeyMacro (KEY_NPAGE);
  DelKeyMacro (KEY_PPAGE);

  free (auswahl->items);
  free (auswahl->resid);
  free (auswahl);

  return result;
}


int DoMenu (char *rname)
{
  MENU *dasmenu = GetMenuResource (rname);
  int   item    = 0, res;

  if (dasmenu == NULL)
    return -1;

  do {
    wattrset (stdscr, A_REVERSE | A_BOLD | COLOR_PAIR (2));
    mvwaddch (stdscr, dasmenu->ypos[item], dasmenu->xpos[item]-1, ' ');
    waddstr  (stdscr, dasmenu->text[item]);
    waddch   (stdscr, ' ');
    wattrset (stdscr, COLOR_PAIR (0));
   
    SetHelpContext (dasmenu->HelpContext);

    res = DoSboxMain (GetSboxResource (dasmenu->submenu[item]),
		      dasmenu->ypos[item]+1, dasmenu->xpos[item]-1,
		      SB_MAYMOVE);

    DelHelpContext ();

    wattrset (stdscr, COLOR_PAIR (2));
    mvwaddch (stdscr, dasmenu->ypos[item], dasmenu->xpos[item]-1, ' ');
    waddstr  (stdscr, dasmenu->text[item]);
    waddch   (stdscr, ' ');
    wattrset (stdscr, COLOR_PAIR (0));

    if (res == SB_LEFT) {
      if (item > 0)
	item--;
      else
	item = dasmenu->anzahl - 1;
    }
    else if (res == SB_RIGHT) {
      if (item + 1 < dasmenu->anzahl)
	item++;
      else
	item = 0;
    }
    else
      break;
  }
  while (TRUE);

  if (res == SB_NONE)
    res = dasmenu->resid[item];
  else if (res == SB_ESCAPE)
    res = -1;
  
  return res;
}

static void _getinfo (struct _lineinfo *res, char *buf, int maxlen, int ltype)
{
  static int flag  = 0;
  char *browse;
  int count = 0;

  if (flag) {
    res->type = AL_BLOCK;
    flag      = 0;
  }
  else
    res->type = ltype;

  while (*buf == ' ')
    buf++;

  if (*buf == '\\') {
    buf++;
    switch (*buf) {
    case 'b':
      res->type = AL_BLOCK;
      flag      = 0;
      break;
    case 'l':
      res->type = AL_LEFT;
      flag      = 0;
      break;
    case 'r':
      res->type = AL_RIGHT;
      flag      = 0;
      break;
    case 'c':
      res->type = AL_CENTER;
      flag      = 0;
      break;
    }
    buf++;
    while (*buf == ' ' || *buf == '\t')
      buf++;
  }

  if (*buf == '\n') {
    res->length = 0;
    res->ptr    = buf + 1;
    if (res->type == AL_BLOCK) {
      res->type = AL_LEFT;
      flag      = 1;
    }
    return;
  }
  if (*buf == '\0') {
    res->length = 0;
    res->ptr    = buf;
    return;
  }

  res->ptr = buf;
  browse   = buf;
  
  do {
    if (*browse == '\n') {
      res->length = count;
      if (res->type == AL_BLOCK) {
	res->type = AL_LEFT;
	flag      = 1;
      }
      return;
    }
    else if (*browse == '\\') {
      res->length = count;
      if (res->type == AL_BLOCK) {
	res->type = AL_LEFT;
	flag      = 1;
      }
      return;
    }
    else if (*browse == '\0') {
      res->length = count;
      if (res->type == AL_BLOCK) 
	res->type = AL_LEFT;
      return;
    }
    count++;
    browse++;
  }
  while (count < maxlen);

  while (*browse != ' ') {
    browse--; count--;
  }
  res->length = count;
  return;
}

void DisplayText (char *rname)
{
  int first=0, known=0, index, last, flag, ym, xm;
  TEXT *tres = GetTextResource (rname);
  char *buf, *tmp, c, **ptr, *sptr;
  struct _lineinfo *lines;
  WINDOW *win, *sub;
  PANEL  *pan, *spa;
  FILE *datei;
  long ccount;

  if (tres == NULL)
    return;

  if (tres->offset == -1) {
    datei = NULL;
    sptr  = tres->filename;
    ptr   = &sptr;
  }
  else {
    if ((datei = fopen (tres->filename, "rt")) == NULL) {
      InfoUser (tres->filename, GetMsgResource ("FileNotFound"));
      return;
    }
    ptr = NULL;
  }
  if ((win = newwin (LINES-6, COLS-20, 3, 10)) == NULL) {
    fclose (datei);
    return;
  }
  if ((pan = new_panel (win)) == NULL) {
    delwin (win);
    fclose (datei);
    return;
  }
  if ((sub = subwin (win, LINES-10, COLS-24, 6, 12)) == NULL) {
    del_panel (pan);
    delwin    (win);
    fclose    (datei);
    return;
  }
  if ((spa = new_panel (sub)) == NULL) {
    del_panel (pan);
    delwin    (sub);
    delwin    (win);
    fclose    (datei);
    return;
  }

  getmaxyx(sub, ym, xm);

  if ((tmp = buf = (char *) malloc (3 * tres->length / 2)) == NULL) {
    del_panel (spa);
    del_panel (pan);
    delwin    (sub);
    delwin    (win);
    fclose    (datei);
    return;
  }
  if ((lines = (struct _lineinfo *) 
       malloc (((2*tres->length)/xm + ym) * sizeof (struct _lineinfo))) == NULL) {
    free      (tmp);
    del_panel (spa);
    del_panel (pan);
    delwin    (sub);
    delwin    (win);
    fclose    (datei);
    return;
  }

  wattrset    (win, COLOR_PAIR (3));
  wfill       (win);
  box         (win, 0, 0);
  whorline    (win, 2);
  wattrset    (win, COLOR_PAIR (4));
  mvwaddstr   (win, 1, 2, tres->title);
  wattrset    (sub, COLOR_PAIR (2));
  SetKeyAlias (KEY_UP,   'k');
  SetKeyAlias (KEY_DOWN, 'j');

  update_panels ();
  doupdate      ();

  if (datei)
    fseek   (datei, tres->offset, SEEK_SET);

  flag = 0;
  for (ccount = 0L; ccount < tres->length; ccount++) {
    *buf = freadc (datei, ptr);

    if (*buf == '\\') {
      *buf = freadc (datei, ptr); ccount++; }

    if (*buf == '\n') {
      if (flag) {
	buf--;
	*buf++ = '\n';
      }
      else {
	*buf++ = ' ';
	flag   = 1;
      }
    }
    else {
      buf++;
      flag = 0;
    }
  }
  if (datei)
    fclose  (datei);

  *buf = '\0'; buf = tmp;

  _getinfo (lines, buf, xm, AL_BLOCK);
  buf = lines->ptr + lines->length; known++;

  while (*buf) {
    _getinfo (lines+known, buf, xm, (lines+known-1)->type);
    buf = (lines+known)->ptr + (lines+known)->length; known++;
  }
  last = known;
  (lines+last)->ptr    = "";
  (lines+last)->length = -1;
  (lines+last)->type   = AL_LEFT; last++; known++;
  (lines+last)->ptr    = GetMsgResource ("PressKey");
  (lines+last)->length = -1;
  (lines+last)->type   = AL_RIGHT;

  SetHelpContext (NULL);
/*
 * Verkraftet seltsamerweise keinen rekursiven Aufruf!
 *
 *  SetHelpContext (tres->HelpContext);
 **/

  do {
    wfill  (sub);
    for (index = 0; index < ym; index++) {
      if (first+index <= last) {
	if ((lines+first+index)->length != -1) {
	  c = *((lines+first+index)->ptr + (lines+first+index)->length);
	  *((lines+first+index)->ptr + (lines+first+index)->length) = '\0';
	}
	woutline (sub, (lines+first+index)->ptr, index, (lines+first+index)->type);

	if ((lines+first+index)->length != -1) 
	  *((lines+first+index)->ptr + (lines+first+index)->length) = c;
      }
    }
    flag = 0;

    do {
      index = InputChar ();
      switch (index) {
      case 'k':
      case KEY_UP:
	if (first > 0) {
	  first--;
	  flag = 1;
	}
	break;
      case 'j':
      case KEY_DOWN:
	if (first+ym <= last) {
	  first++;
	  flag = 1;
	}
	break;
      case '0':
      case KEY_HOME:
	if (first != 0) {
	  first = 0;
	  flag  = 1;
	}
	break;
      case CTRL('F'):
      case KEY_NPAGE:
	if (first+ym-1 <= last) {
	  first += ym-1;
	  flag   = 1;
	}
	break;
      case CTRL('B'):
      case KEY_PPAGE:
	if (first-ym+1 >= 0) {
	  first -= ym-1;
	  flag   = 1;
	}
	else {
	  if (first != 0)
	    flag = 1;
	  first  = 0;
	}
	break;
      case '$':
      case KEY_END:
	if (last >= ym-1) {
	  first = last - ym + 1;
	  flag  = 1;
	}
	break;
      case 0x1b:
      case '\n':
      case KEY_F(0):
	flag = 1;
	break;
      }
    }
    while (!flag);
  }
  while (index != '\n' && index != 0x1b && index != KEY_F(0));

  DelHelpContext ();

  DelKeyAlias (KEY_DOWN);
  DelKeyAlias (KEY_UP);
  del_panel   (spa);
  del_panel   (pan);
  delwin      (sub);
  delwin      (win);
  free        (tmp);
  free        (lines);
}

PCBAR *InitBar (int maxval, int ypos, int xpos, int length)
{
  PCBAR *result;

  switch (-ypos) {
  case AL_LEFT:  ypos=1; break;
  case AL_RIGHT: ypos= LINES-3; break;
  case AL_CENTER:ypos=(LINES-3) / 2; break;
  default:       break;
  }
  switch (-xpos) {
  case AL_LEFT:  xpos=1; break;
  case AL_RIGHT: xpos= COLS-length-2; break;
  case AL_CENTER:xpos=(COLS-length-2) / 2; break;
  default:       break;
  }

  if ((result = (PCBAR *) malloc (sizeof (PCBAR))) == NULL)
    return NULL;
  if ((result->win = newwin (3, length+2, ypos-1, xpos-1)) == NULL) {
    free (result);
    return NULL;
  }
  if ((result->pan = new_panel (result->win)) == NULL) {
    delwin (result->win);
    free   (result);
    return NULL;
  }
  wattrset  (result->win, COLOR_PAIR (0) | A_REVERSE);
  wfill     (result->win);
  wattrset  (result->win, COLOR_PAIR (3));
  box       (result->win, 0, 0);
  wattrset  (result->win, COLOR_PAIR (0) | A_REVERSE | A_BOLD);

  result->length = length;
  result->value  = 0;
  result->maxval = (maxval <= 0) ? 1 : maxval;

  return result;
}

void SetBar (PCBAR *bar, int value, int maxval)
{
  int flag=0, anz, count;

  if (bar == NULL)
    return;

  if (maxval > 0) {
    bar->maxval = maxval;
    flag        = 1;
  }

  bar->value = value;
  anz = (bar->length * bar->value) / bar->maxval;

  for (count = 0; count < anz; count++)
    mvwaddch (bar->win, 1, count+1, '#');

  if (flag)
    for (count = anz; count < bar->length; count++)
      mvwaddch (bar->win, 1, count+1, ' ');

  mvwprintw (bar->win, 1, bar->length/2 - 2, " %02d %% ", 100 * bar->value / bar->maxval);

  return;
}

void DelBar (PCBAR *bar)
{
  if (bar != NULL) {
    del_panel (bar->pan);
    delwin    (bar->win);
    free      (bar);
  }
}

int InputText (WINDOW *win, char *buf, int ypos, int xpos, int len, int type)
{
  int xptr=0, c, flag, index;
  int modified = 0, x, y;
  
  flag  = 1; 
  index = strlen (buf);

  if (index < len) {
    memset (buf+index, ' ', len - index);
    *(buf+len) = '\0';
  }

  do {
    if (flag) {
      flag = 0;
      mvwaddstr (win, ypos, xpos, buf);
    }

    update_panels ();
    doupdate      ();

    getbegyx (win, y, x);
    wmove    (win, ypos, xpos+xptr);
    wmove    (stdscr, y+ypos, x+xpos+xptr);

    c = InputChar ();

    if (c < KEY_MIN &&
	((isalpha (c) && type & IT_ALPHA) || 
	 (isdigit (c) && type & IT_NUMERIC) ||
	 (ispunct (c) && type & IT_CHAR) ||
	 (c == ' '    && type & IT_SPACE))) {

      modified = 1;

      if (type & IT_UPCASE && islower (c))
	c = toupper (c);
      else if (type & IT_DCASE && isupper (c))
	c = tolower (c);
      
      for (index=len-1; index>xptr; index--)
	*(buf+index) = *(buf+index-1);
      *(buf+xptr) = (char) c;
      flag        = 1;
      if (++xptr == len && type & IT_TABEXIT) {
	c = '\t';
	break;
      }
    }
    else if ((c == '\t' || c == KEY_BTAB) && type & IT_TABEXIT)
      break;
    else if (c == '\n' || c == '\x1b' || c == KEY_F(0) || c == CTRL('X'))
      break;
    else if ((c >= KEY_F(0) && c <= KEY_F(63)) && type & IT_FKEYEXIT)
      break;
    else if (c == KEY_F(6) || c == KEY_F(8))
      break;
    else 
      switch (c) {
      case KEY_RIGHT:
	if (xptr+1 < len)
	  xptr++;
	break;
      case KEY_LEFT:
	if (xptr > 0)
	  xptr--;
	break;
      case KEY_BEG:
      case KEY_HOME:
	xptr = 0;
	break;
      case KEY_END:
	xptr = len-1;
	while (xptr && *(buf+xptr)==' ')
	  xptr--;
	if (*(buf+xptr) != ' ' && xptr < len-1)
	  xptr++;
	break;
      case KEY_BACKSPACE:
      case '\x08':
	if (xptr > 0) {
	  for (index=xptr; index<len; index++)
	    *(buf+index-1) = *(buf+index);
	  *(buf+len-1)     = ' ';
	  flag = modified  = 1;
	  xptr--;
	}
	break;
      case KEY_DC:
      case '\x7f':
	for (index=xptr; index<len-1; index++)
	  *(buf+index)  = *(buf+index+1);
	*(buf+len-1)    = ' ';
	flag = modified = 1;
	break;
      case CTRL('E'):
	for (index=xptr; index<len; index++)
	  *(buf+index)  = ' ';
	flag = modified = 1;
	break;
      case CTRL('U'):
	for (index=0; index<len; index++)
	  *(buf+index)  = ' ';
	flag = modified = 1;
	xptr = 0;
	break;
      }
  }
  while (TRUE);

  mvwaddstr (win, ypos, xpos, buf);

  if ((c >= KEY_F(0) && c <= KEY_F(63)) && type & IT_FKEYEXIT)
    return c;

  switch (c) {
  case '\t':
    return IT_TAB;  /* | ((type & IT_ACTLEAVE)?((modified)?IT_MODIFIED:0):0); */
  case KEY_BTAB:
    return IT_BTAB; /* | ((type & IT_ACTLEAVE)?((modified)?IT_MODIFIED:0):0); */
  case CTRL('X'):
  case KEY_F(0):
  case '\x1b':
    return IT_ESCAPE;
  case KEY_F(6):
    return IT_DEFAULT;
  case KEY_F(8):
    return IT_CLFORM;
  default:
    return IT_OK;   /* | ((type & IT_ACTLEAVE)?((modified)?IT_MODIFIED:0):0); */
  }
  return IT_OK; /* Not to be reached */
}

FINFO *InitForm (char *res)
{
  FORM *this = GetFormResource (res), *copy;
  int index, yf, count, ypos, xpos;
  WINDOW *win, *subw;
  PANEL *pan, *subp;
  FINFO *result;

  if (this == NULL)
    return NULL;
  if (this->title == NULL || this->title[0] == '\0')
    yf = 2;
  else
    yf = 4;

  xpos = this->xpos;
  ypos = this->ypos;

  switch (-ypos) {
  case AL_LEFT:  ypos=yf/2; break;
  case AL_RIGHT: ypos= LINES-this->yres-yf/2; break;
  case AL_CENTER:ypos=(LINES-this->yres) / 2; break;
  default:       break;
  }
  switch (-xpos) {
  case AL_LEFT:  xpos=2; break;
  case AL_RIGHT: xpos= COLS-this->xres-2; break;
  case AL_CENTER:xpos=(COLS-this->xres) / 2; break;
  default:       break;
  }

  if ((win = newwin (this->yres+yf,   this->xres+4, ypos-yf/2, xpos-2)) == NULL)
    return NULL;
  if ((pan = new_panel (win)) == NULL) {
    delwin (win);
    return NULL;
  }
  if ((subw = subwin (win, this->yres, this->xres, ypos+(yf-2)/2, xpos)) == NULL) {
    del_panel (pan);
    delwin    (win);
    return NULL;
  }
  if ((subp = new_panel (subw)) == NULL) {
    del_panel (pan);
    delwin    (subw);
    delwin    (win);
    return NULL;
  }
  if ((result = (FINFO *) malloc (sizeof (FINFO))) == NULL) {
    del_panel (subp);
    del_panel (pan);
    delwin    (subw);
    delwin    (win);
    return NULL;
  }
  if ((copy = (FORM *) malloc (sizeof (FORM))) == NULL) {
    free      (result);
    del_panel (subp);
    del_panel (pan);
    delwin    (subw);
    delwin    (win);
    return NULL;
  }

  memcpy (copy, this, sizeof (FORM));

  if ((copy->buffer = (char **) malloc (this->anzahl * sizeof (char *))) == NULL) {
    free      (copy);
    del_panel (subp);
    del_panel (pan);
    delwin    (subw);
    delwin    (win);
    return NULL;
  }

  wattrset  (win,  COLOR_PAIR (2));
  wattrset  (subw, COLOR_PAIR (2));
  wfill     (win);
  wfill     (subw);
  wattrset  (win,  COLOR_PAIR (3));
  box       (win, 0, 0);

  if (copy->title != NULL)
    whorline(win, 2);

  wattrset  (win,  COLOR_PAIR (4));

  if (copy->title != NULL)
    mvwaddstr (win, 1, 2, copy->title);

  for (index=0; index < copy->anzahl; index++) {
    if (copy->label[index] != NULL) {
      wattrset  (subw, COLOR_PAIR (2));
      mvwaddstr (subw, copy->lypos[index], copy->lxpos[index], copy->label[index]);
    }

    if (copy->flen[index] > 0) {
      if ((copy->buffer[index] = (char *) malloc (sizeof (char) * (copy->flen[index]+1))) == NULL)
	copy->flen[index] = 0;
      else {
	memset (copy->buffer[index], ' ', copy->flen[index]);
	*(copy->buffer[index]+copy->flen[index]) = '\0';
      }
    }
    else
      copy->buffer[index] = NULL;

    if (copy->flen[index] != 0) {
      if (copy->ftype[index] & IT_INACTIVE)
	wattrset  (subw, COLOR_PAIR (2));
      else if (copy->ftype[index] & IT_DOWNLIGHT)
	wattrset (subw, COLOR_PAIR (3));
      else 
	wattrset (subw, COLOR_PAIR (0) | A_REVERSE | A_BOLD);
      wmove      (subw, copy->fypos[index], copy->fxpos[index]);
      
      for (count=0; count < copy->flen[index]; count++)
	waddch    (subw, ' ');
    }
  }
  
  hide_panel (subp);
  hide_panel (pan);

  result->eflag = 0;
  result->pan   = pan;
  result->sub   = subp;
  result->form  = copy;

  return result;
}

void PutForm (FINFO *info, int resid, char *text)
{
  int index=0, len;

  if (info == NULL)
    return;
  if (text == NULL)
    text = "";

  len = strlen (text);

  while (index < info->form->anzahl) {
    if (info->form->resid[index] == resid)
      break;
    else
      index++;
  }

  if (index < info->form->anzahl) {
    if (len > info->form->flen[index])
      len = info->form->flen[index];

    memcpy (info->form->buffer[index], text, len);
    memset (info->form->buffer[index]+len, ' ', info->form->flen[index]-len);
    *(info->form->buffer[index]+info->form->flen[index]) = '\0';

    if (info->form->ftype[index] & IT_INACTIVE)
      wattrset(panel_window (info->sub), COLOR_PAIR (2));
    else if (info->form->ftype[index] & IT_DOWNLIGHT)
      wattrset(panel_window (info->sub), COLOR_PAIR (3));
    else
      wattrset(panel_window (info->sub), COLOR_PAIR (0) | A_REVERSE | A_BOLD);

    mvwaddstr (panel_window (info->sub),
	       info->form->fypos[index],
	       info->form->fxpos[index],
	       info->form->buffer[index]);
  }
}

void PutFormDefaults (FINFO *info)
{
  int index;

  for (index=0; index<info->form->anzahl; index++)
    if (info->form->flen[index])
      PutForm (info, info->form->resid[index],
	       (info->form->stand[index]) ? info->form->stand[index] : "");
}

char *GetForm (FINFO *info, int resid)
{
  int index=0;

  if (info == NULL)
    return NULL;
  while (index < info->form->anzahl) {
    if (info->form->resid[index] == resid)
      return info->form->buffer[index];
    else
      index++;
  }
  return NULL;
}

int QueryForm (FINFO *info)
{
  int wobinich, result, index;
  int wowarich, hflag = 0;

  if (info == NULL)
    return IT_ESCAPE;

  if (result = info->eflag) {
    info->eflag = 0;
    return (result & 0xbfff);
  }

  if (panel_hidden (info->pan)) {
    DisplayForm (info);
    hflag = 1;
  }

  wobinich = info->form->field;

  while (info->form->flen [wobinich] == 0 ||
	 info->form->ftype[wobinich] & IT_INACTIVE) {
    if (++wobinich >= info->form->anzahl)
      return IT_ESCAPE;
  }

  SetHelpContext (info->form->HelpContext);
  SetKeyAlias    (KEY_UP,   KEY_BTAB);
  SetKeyAlias    (KEY_DOWN, '\t');

  do {
    if (info->form->ftype[wobinich] & IT_DOWNLIGHT)
      wattrset (panel_window (info->sub), COLOR_PAIR (3));
    else
      wattrset (panel_window (info->sub), COLOR_PAIR (0) | A_REVERSE | A_BOLD);

    result = InputText (panel_window (info->sub),
			info->form->buffer[wobinich],
			info->form->fypos[wobinich],
			info->form->fxpos[wobinich],
			info->form->flen [wobinich],
			info->form->ftype[wobinich] | IT_TABEXIT);

    if (result == IT_ESCAPE)
      break;
    else if (result == IT_OK) {
      if (info->form->ftype[wobinich] & IT_ACTLEAVE) {
	result = - info->form->resid[wobinich];
	info->eflag = IT_OK | 0x4000;
      }
      break;
    }
    else if (result == IT_TAB) {
      wowarich = wobinich;

      if (++wobinich >= info->form->anzahl)
	wobinich = 0;
      while (info->form->flen [wobinich] == 0 ||
	     info->form->ftype[wobinich] & IT_INACTIVE) {
	if (++wobinich >= info->form->anzahl)
	  wobinich = 0;
      }
      if (info->form->ftype[wowarich] & IT_ACTLEAVE) {
	result = - info->form->resid[wowarich];
	break;
      }
    }
    else if (result == IT_BTAB) {
      wowarich = wobinich;

      if (--wobinich < 0)
	wobinich = info->form->anzahl - 1;
      while (info->form->flen[wobinich] == 0) {
	if (--wobinich < 0)
	  wobinich = info->form->anzahl - 1;
      }
      if (info->form->ftype[wowarich] & IT_ACTLEAVE) {
	result = - info->form->resid[wowarich];
	break;
      }
    }
    else if (result >= KEY_F(0) && result <= KEY_F(63)) {
      if (info->form->ftype[wobinich] & IT_ACTLEAVE) {
	info->eflag = result | 0x4000;
	result = - info->form->resid[wobinich];
      }
      break;
    }
    else if (result == IT_DEFAULT) {
      for (index=0; index<info->form->anzahl; index++)
	if (info->form->flen[index])
	  PutForm (info, info->form->resid[index],
		   (info->form->stand[index]) ? info->form->stand[index] : "");
    }
    else if (result == IT_CLFORM) {
      for (index=0; index<info->form->anzahl; index++)
	if (info->form->flen[index])
	  PutForm (info, info->form->resid[index], NULL);
    }
  }
  while (TRUE);

  DelKeyAlias    (KEY_DOWN);
  DelKeyAlias    (KEY_UP);
  DelHelpContext ();

  if (hflag)
    HideForm (info);

  info->form->field = wobinich;
  return result;
}

void SetFormCursor (FINFO *info, int resid)
{
  int index=0;

  if (info == NULL)
    return;

  while (index < info->form->anzahl) {
    if ((info->form->resid[index] == resid) &&
	(!(info->form->ftype[index] & IT_INACTIVE))) {
      info->form->field = index;
      break;
    }
    else
      index++;
  }
}

int GetFormCursor (FINFO *info)
{
  if (info == NULL)
    return 0;

  return info->form->resid[info->form->field];
}

void ClearForm (FINFO *info)
{
  int index;

  if (info == NULL)
    return;

  for (index=0; index<info->form->anzahl; index++)
    if (info->form->flen[index])
      PutForm (info, info->form->resid[index], NULL);
}

int GetFormFields (FINFO *info)
{
  if (info == NULL)
    return;

  return info->form->anzahl;
}

void DeleteForm (FINFO *info)
{
  WINDOW *win, *sub;
  int index;

  if (info == NULL)
    return;

  win = panel_window (info->pan);
  sub = panel_window (info->sub);

  for (index=0; index < info->form->anzahl; index++)
    if (info->form->buffer[index] != NULL)
      free (info->form->buffer[index]);

  del_panel (info->sub);
  del_panel (info->pan);
  delwin    (sub);
  delwin    (win);

  free      (info->form);
  free      (info);
}

void MoveForm (FINFO *this, int yneu, int xneu)
{
  int yf;

  if (yneu < 0 || xneu < 0 || this == NULL)
    return;

  if (this->form->title == NULL)
    yf = 2;
  else
    yf = 4;
  
  move_panel (this->pan, yneu-yf/2, xneu-2);
  move_panel (this->sub, yneu+(yf-2)/2, xneu);

  this->form->ypos = yneu;
  this->form->xpos = xneu;
}

void DisplayForm (FINFO *form)
{
  if (form != NULL) {
    if (panel_hidden (form->pan)) {
      show_panel (form->pan);
      show_panel (form->sub);
    }
    else {
      top_panel  (form->pan);
      top_panel  (form->sub);
    }
  }
}

void HideForm (FINFO *form)
{
  if (form != NULL && !panel_hidden (form->pan)) {
    hide_panel (form->sub);
    hide_panel (form->pan);
  }
}

/*
 * Hardcopy in eine Datei
 **/

void ScreenToFile (FILE *datei)
{
  int y, x;

  for (y=0; y<LINES; y++) {
    for (x=0; x<COLS; x++)
      fputc (simplechar (mvwinch (curscr, y, x)), datei);
    fputc ('\n', datei);
  }
  fputc ('\n', datei);
}

/*
 * Signalhandler fuer Control-C, Break!
 **/

void SigSetBreak (int sig)
{
  FINFO *bform;

/*
 * Signalhandler blockiert?
 * Ja: Neu installieren und Ende
 **/
  if (BlockBreak) {
#ifdef SIGACTION
    act.sa_handler = SigSetBreak;
    sigaction (SIGINT, &act, NULL);
#else
    signal (SIGINT, SigSetBreak);
#endif
    return;
  }

/*
 * Wurde bereits ein Control-C eingefangen?
 * Wenn nein -> Abfragebildschirm einblenden
 **/
  if (!HaveABreak ()) {
    if ((bform = InitForm ("ControlCForm")) != NULL) {
      PutForm (bform, 700, GetMsgResource ("YesText"));
      if (QueryForm (bform) == IT_OK) {
	if (strncmp (GetForm (bform, 700),
		     GetMsgResource ("YesText"),
		     strlen (GetMsgResource ("YesText"))) == 0) {
	  if ((AbbruchMeld = InitForm ("IWannaBreak")) != NULL)
	    DisplayForm (AbbruchMeld);
	  isBreak++;
	}
      }
      DeleteForm (bform);
      update_panels ();
      doupdate      ();
    }
  }
  else
    isBreak++;

/*
 * Falls isBreak auch nach 9 Control-C's noch nicht abgefangen wurde,
 * dann breche bei Nr. 10 ab
 **/
  if (isBreak < 9) {
#ifdef SIGACTION
    act.sa_handler = SigSetBreak;
    sigaction (SIGINT, &act, NULL);
#else
    signal (SIGINT, SigSetBreak);
#endif
  }
  else {
#ifdef SIGACTION
    act.sa_handler = SIG_DFL;
    sigaction (SIGINT, &act, NULL);
#else
    signal (SIGINT, SIG_DFL);
#endif
  }
}

/*
 * Soll abgebrochen werden?
 **/

int HaveABreak (void)
{
  return isBreak;
}

/*
 * Control-C wurde bearbeitet
 **/
void BreakProcessed (void)
{
  DeleteForm (AbbruchMeld);
  AbbruchMeld = NULL;
  isBreak     = 0;
}

/*
 * Signalhandler einbinden
 **/
void InstallCtrlC (void)
{
#ifdef SIGACTION
  act.sa_handler = SigSetBreak;
  sigaction (SIGINT, &act, NULL);
#else
  signal (SIGINT, SigSetBreak);
#endif
}

void DeInstallCtrlC (void)
{
#ifdef SIGACTION
  act.sa_handler = SIG_IGN;
  sigaction (SIGINT, &act, NULL);
#else
  signal (SIGINT, SigSetBreak);
#endif
}
