/*****************************************************************************
******************************************************************************
*****                                                                    *****
*****     resource.h                                                     *****
*****                   Bildschirmfunktionen                             *****
*****                                                                    *****
*****     ??.03.93      erstellt                                   FP    *****
*****                                                                    *****
******************************************************************************
*****************************************************************************/

#ifndef __RESOURCE_H__
#define __RESOURCE_H__

/*
 * Define some possibly undefined Keys to hopefully impossible values
 * But we may need KEY_MIN, the minimum 'special key', this is usually 
 * 0400 (oct).
 */

#ifndef KEY_MIN
#  define KEY_MIN	0400
#endif
#ifndef KEY_HELP
#  define KEY_HELP	01000
#endif
#ifndef KEY_END
#  define KEY_END	01001
#endif
#ifndef KEY_BEG
#  define KEY_BEG	01002
#endif

/*
 * Some systems don't have these macros. getmaxyx gets the window size,
 * getbegyx the absolute position of a window's upper left corner.
 * Remember that these macros take the variables itself as parameter,
 * and not their addresses.
 */

#ifndef getmaxyx
#  define getmaxyx(w,y,x)	{y=w->_maxy;x=w->_maxx;}
#endif
#ifndef getbegyx
#  define getbegyx(w,y,x)	{y=w->_begy;x=w->_begx;}
#endif

/*
 * Do we have a proper ACS_ Set? If not, choose one.
 * ACS_SSSS and ACS_PLUS are hopefully proper representants
 * of all half-graphical characters
 */

#ifndef ACS_SSSS			/* Search for 'standard-style' */
#  ifndef ACS_PLUS			/* AND  'human-readable-style' */
#    define ACS_SSSS	'+'
#    define ACS_BSSB	'+'		/* We don't have any graphical */
#    define ACS_SSBB	'+'		/* characters -- use standard  */ 
#    define ACS_BBSS	'+'		/* chars                       */
#    define ACS_SBBS	'+'
#    define ACS_SBSS	'+'
#    define ACS_SSSB	'+'
#    define ACS_SSBS	'+'
#    define ACS_BSSS	'+'
#    define ACS_BSBS	'-'
#    define ACS_SBSB	'|'
#  else
#    define ACS_BSSB	ACS_ULCORNER	/* Use 'human-readable' to setup */
#    define ACS_SSBB	ACS_LLCORNER	/* 'standard-style'              */
#    define ACS_BBSS	ACS_URCORNER
#    define ACS_SBBS	ACS_LRCORNER
#    define ACS_SBSS	ACS_RTEE
#    define ACS_SSSB	ACS_LTEE
#    define ACS_SSBS	ACS_BTEE
#    define ACS_BSSS	ACS_TTEE
#    define ACS_BSBS	ACS_HLINE
#    define ACS_SBSB	ACS_VLINE
#    define ACS_SSSS	ACS_PLUS
#  endif
#endif
					/* We still might not have 'human' */
#ifndef ACS_PLUS			/* so set 'em up                   */
#  define ACS_ULCORNER 	ACS_BSSB
#  define ACS_LLCORNER 	ACS_SSBB
#  define ACS_URCORNER 	ACS_BBSS
#  define ACS_LRCORNER 	ACS_SBBS
#  define ACS_RTEE	ACS_SBSS
#  define ACS_LTEE	ACS_SSSB
#  define ACS_BTEE	ACS_SSBS
#  define ACS_TTEE	ACS_BSSS
#  define ACS_HLINE	ACS_BSBS
#  define ACS_VLINE	ACS_SBSB
#  define ACS_PLUS	ACS_SSSS
#endif

#ifndef COLOR_WHITE			/* If we don't have colors ... */
#  define start_color()	(ERR)		/* Setup dummy values          */
#  define COLOR_PAIR(n)	(0)
#  define _NOCOLORS_
#endif

#define FGETCC(f,p)	((f)?(fgetc(f)):((**(p))?(*((*(p))++)):(**(p))))
#define FEOFC(f,p)	((f)?(feof(f)):((**(p))?(FALSE):(TRUE)))
#define UNGETCC(c,f,p)	((f)?(ungetc(c,f)):((int)((*(p))--)))

#define SB_STANDARD	0
#define SB_TITLED	1
#define SB_MAYMOVE	2

#define AL_LEFT		1
#define AL_RIGHT	2
#define AL_CENTER	3
#define AL_BLOCK	4

#define SB_ESCAPE	1
#define SB_LEFT		2
#define SB_RIGHT	3
#define SB_NONE		4

#define IT_ALPHA	0x0001
#define IT_NUMERIC	0x0002
#define IT_CHAR		0x0004
#define IT_UPCASE	0x0008
#define IT_DCASE	0x0010
#define IT_TABEXIT	0x0020
#define IT_SPACE	0x0040
#define IT_INACTIVE	0x0080
#define IT_YESNO	0x0100
#define IT_DOWNLIGHT	0x0200
#define IT_FKEYEXIT	0x0400
#define IT_ACTENTER	0x0800
#define IT_ACTLEAVE	0x1000
#define IT_ALNUM	(IT_ALPHA | IT_NUMERIC)

#define IT_OK		0
#define IT_ESCAPE	1
#define IT_TAB		2
#define IT_BTAB		3
#define IT_DEFAULT	4
#define IT_CLFORM	5
#define IT_MODIFIED     0x2000

#define AI_UP 		1
#define AI_RIGHT	2
#define AI_DOWN		4
#define AI_LEFT		8

#ifndef CTRL
#define CTRL(c)		((c) & 037)
#endif

/******
******* Resources
******/

#define RES_NONE	(-1)
#define RES_SBOX	1
#define RES_MENU	2
#define RES_TEXT	3
#define RES_MSG		4
#define RES_FORM	5
#define RES_TVEC	6

typedef struct _sbox {
  char *rname;
  char *title;
  int anzahl;
  int ypos, xpos;
  int ysize, xsize;
  int *resid;
  char **items;
  char *HelpContext;
} SBOX;

typedef struct _menu {
  char *rname;
  int   anzahl;
  char *text[20];
  int   resid[20];
  int   ypos[20], xpos[20];
  char *submenu[20];
  char *HelpContext;
} MENU;

typedef struct _text {
  char *rname;
  char *title;
  long offset;
  long length;
  char *filename;
  char *HelpContext;
} TEXT;

typedef struct _message {
  char *rname;
  char *text;
  char *HelpContext;
} MESSAGE;

typedef struct _form {
  char *rname;
  char *title;
  int yres, xres, ypos, xpos;
  int anzahl, field;
  int *lypos, *lxpos;
  int *fypos, *fxpos;
  int *flen, *resid, *ftype;
  char **label;
  char **buffer;
  char **stand;
  char *HelpContext;
} FORM;

typedef struct _tvector {
  char *rname;
  int anzahl;
  char **elements;
  char *HelpContext;
} TVECTOR;

typedef struct _resource {
  int type;
  char *rname;
  union {
    SBOX *sbox;
    MENU *menu;
    TEXT *text;
    FORM *form;
    MESSAGE *msg;
    TVECTOR *tvec;
  } data;
  struct _resource *LEFT;
  struct _resource *RIGHT;
} RESOURCE;

typedef struct _fifo {
  int eflag;
  PANEL *pan;
  PANEL *sub;
  FORM *form;
} FINFO;

typedef struct _pcbar {
  long length;
  long value;
  long maxval;
  WINDOW *win;
  PANEL *pan;
} PCBAR;

extern char accpref[];

int		minical			(char *);
char *		calcstr			(char *, char *);
char *		FGETSC			(FILE *, char **, char *);
char 		freadc 			(FILE *, char **);
int		freadi			(FILE *, char **);
char *		freads			(FILE *, char **);
void		fungetc			(int c);

char *		ShortString		(char *);
int		IsStringEmpty		(char *);

RESOURCE *	LoadResource 		(char *, char *);
void 		KillResource 		(RESOURCE *);
void 		AddMoreResources 	(RESOURCE *);

RESOURCE *	GetResource		(char *, int);

SBOX *		GetSboxResource 	(char *);
MENU *		GetMenuResource 	(char *);
TEXT *		GetTextResource 	(char *);
char *		GetMsgResource		(char *);
char **		GetTvecResource		(char *);

int  		StartCurses 		(char *, char *, int);
void		EndCurses		(void);
int		DoResize		(int, int);
int 		GetWinSize 		(int *, int *);
int 		HavetoResize 		(void);

void		ScreenToFile		(FILE *);

int		HaveABreak		(void);
void		BreakProcessed		(void);
void		InstallCtrlC		(void);
void		DeInstallCtrlC		(void);

int		InputChar		(void);

void 		wfill 			(WINDOW *);
void 		wfillrgn 		(WINDOW *, int, int, int, int);
int 		acs2int 		(chtype);
chtype 		int2acs 		(int);
int		simplechar		(int);
void 		wladdch 		(WINDOW *, int, int, int);
void 		wbox 			(WINDOW *, int, int, int, int);
void 		wline 			(WINDOW *, int, int, int, int);
void 		whorline 		(WINDOW *, int);
void 		woutline 		(WINDOW *, char *, int, int);
void		wprintstr		(WINDOW *, char *, int, int);
void		RedrawScreen		(void);
int 		InputText 		(WINDOW *, char *, int, int, int, int);

void		DisplayHelp		(void);
void		SetHelpContext		(char *);
void		DelHelpContext		(void);
void 		SetKeyAlias		(int, int);
void		DelKeyAlias		(int);
void 		SetKeyHook 		(int, void (*function) (void));
void		DelKeyHook		(int);
void		SetKeyMacro		(int, int *);
void		DelKeyMacro		(int);

void 		InfoUser		(char *, char *);
int		DoSbox			(char *);
int 		DoSboxMain 		(SBOX *, int, int, int);
int		DynSelectBox		(int, char *, char *, char *, char *, char **);
void 		InitMenu 		(char *);
int		DoMenu 			(char *);
void 		DisplayText 		(char *);

PCBAR *		InitBar			(int, int, int, int);
void		SetBar			(PCBAR *, int, int);
void		DelBar			(PCBAR *);

FINFO *		InitForm		(char *);
void 		PutForm 		(FINFO *, int, char *);
char *		GetForm 		(FINFO *, int);
void		PutFormDefaults		(FINFO *);
int 		QueryForm 		(FINFO *);
void		SetFormCursor		(FINFO *, int);
int		GetFormCursor		(FINFO *);
int		GetFormFields		(FINFO *);
void		DisplayForm		(FINFO *);
void 		MoveForm 		(FINFO *, int, int);
void		HideForm		(FINFO *);
void		ClearForm		(FINFO *);
void 		DeleteForm 		(FINFO *);

#endif
