/*****************************************************************************
******************************************************************************
*****                                                                    *****
*****     sweep.c                                                        *****
*****                   Keine Ahnung was das sein soll --- halt wieder   *****
*****                   ein schoenes undokumentiertes Feature ...!       *****
*****                                                                    *****
*****     26.08.93      erstellt                                   FP    *****
*****                                                                    *****
******************************************************************************
*****************************************************************************/

#include <curses.h>

#ifdef NOPANEL
#include <nopanel.h>
#else
#include <panel.h>
#endif

#include <time.h>
#include "resource.h"

#define NOMINE		0x00
#define ISMINE		0x80
#define FREE		0x01
#define MINEGUESS	0x02
#define QUESTION	0x04

#define MS_OK		0x00
#define MS_UP		0x01
#define MS_DOWN		0x02
#define MS_LEFT		0x04
#define MS_RIGHT	0x08
#define MS_EXIT		0x20
#define MS_TAG		0x40
#define MS_BLOWUP	0x80

typedef struct {
  int ysize, xsize;
  int mines, guess, found;
  int closed;
  char *data;
} MINEFIELD;

int MS_Size[3][3] = {
  { 10, 10,  10 },
  { 15, 30,  70 },
  { 20, 50, 150 }
};

char MS_PeekField (MINEFIELD *field, int ypos, int xpos)
{
  return *(field->data + ypos*field->xsize + xpos);
}

char MS_PokeField (MINEFIELD *field, int ypos, int xpos, int val)
{
  *(field->data + ypos*field->xsize + xpos) = val;
}

int MS_IsMine (MINEFIELD *field, int ypos, int xpos)
{
  if (ypos < 0 || xpos < 0 || ypos >= field->ysize || xpos >= field->xsize)
    return 0;

  return (int) MS_PeekField (field, ypos, xpos) & ISMINE;
}

int MS_FieldMines (MINEFIELD *field, int ypos, int xpos)
{
  int count = 0;

  if (MS_IsMine (field, ypos-1, xpos-1)) count++;
  if (MS_IsMine (field, ypos,   xpos-1)) count++;
  if (MS_IsMine (field, ypos+1, xpos-1)) count++;
  if (MS_IsMine (field, ypos-1, xpos  )) count++;
  if (MS_IsMine (field, ypos+1, xpos  )) count++;
  if (MS_IsMine (field, ypos-1, xpos+1)) count++;
  if (MS_IsMine (field, ypos,   xpos+1)) count++;
  if (MS_IsMine (field, ypos+1, xpos+1)) count++;

  return count;
}

MINEFIELD *MS_InitField (int ysize, int xsize, int mines)
{
  MINEFIELD *result;
  int yrand, xrand;

  if ((result = (MINEFIELD *) malloc (sizeof (MINEFIELD))) == NULL)
    return NULL;
  if ((result->data = (char *) malloc (ysize*xsize)) == NULL) {
    free (result);
    return NULL;
  }
  memset (result->data, NOMINE, ysize*xsize);

  result->ysize = ysize;
  result->xsize = xsize;
  result->mines = mines;
  result->found = 0;
  result->guess = 0;
  result->closed= xsize * ysize;

  srand ((unsigned int) ((1234L * time(NULL) + time(NULL)) % 65536));

  while (mines > 0) {
    yrand = rand () % ysize;
    xrand = rand () % xsize;

    if (!MS_IsMine (result, yrand, xrand)) {
      MS_PokeField (result, yrand, xrand, ISMINE); mines--; }
  }

  return result;
}

void MS_KillField (MINEFIELD *field)
{
  free (field->data);
  free (field);
}

int MS_OpenUp (MINEFIELD *field, int ypos, int xpos)
{
  int res = 0;

  if (ypos < 0 || xpos < 0 || ypos >= field->ysize || xpos >= field->xsize)
    return 0;

  if (MS_IsMine (field, ypos, xpos))
    return 0;
  if (MS_PeekField (field, ypos, xpos) & MINEGUESS)
    return 0;
  if (MS_PeekField (field, ypos, xpos) == FREE)
    return 0;

  MS_PokeField (field, ypos, xpos, FREE); res++;

  if (MS_FieldMines (field, ypos, xpos) == 0) {
    res += MS_OpenUp (field, ypos-1, xpos-1);
    res += MS_OpenUp (field, ypos,   xpos-1);
    res += MS_OpenUp (field, ypos+1, xpos-1);
    res += MS_OpenUp (field, ypos-1, xpos  );
    res += MS_OpenUp (field, ypos+1, xpos  );
    res += MS_OpenUp (field, ypos-1, xpos+1);
    res += MS_OpenUp (field, ypos,   xpos+1);
    res += MS_OpenUp (field, ypos+1, xpos+1);
  }
  return res;
}
  
int MS_DoCursor (MINEFIELD *field, int ypos, int xpos)
{
  int key, res=-1;
  char test;

  while (res == -1) {
    key = InputChar ();
    switch (key) {
    case 'k':
    case KEY_UP:
      if (ypos > 0)
	res = MS_UP;
      break;
    case 'j':
    case KEY_DOWN:
      if (ypos+1 < field->ysize)
	res = MS_DOWN;
      break;
    case 'h':
    case KEY_LEFT:
      if (xpos > 0)
	res = MS_LEFT;
      break;
    case 'l':
    case KEY_RIGHT:
      if (xpos+1 < field->xsize)
	res = MS_RIGHT;
      break;
    case 't':
      test = MS_PeekField (field, ypos, xpos);

      if (test == FREE)
	continue;

      switch (test & (MINEGUESS | QUESTION)) {
      case 0:
	MS_PokeField (field, ypos, xpos, test | MINEGUESS);
	field->guess += 1;
	field->closed-= 1;

	if (MS_IsMine (field, ypos, xpos))
	  field->found += 1;

	res = MS_TAG;
	break;
      case MINEGUESS:
	MS_PokeField (field, ypos, xpos, (test & ~MINEGUESS) | QUESTION);
	field->guess -= 1;
	field->closed+= 1;

	if (MS_IsMine (field, ypos, xpos))
	  field->found -= 1;

	res = MS_TAG;
	break;
      case QUESTION:
	MS_PokeField (field, ypos, xpos, test & ~QUESTION);
	res = MS_TAG;
	break;
      }
      break;
    case ' ':
      if (MS_PeekField (field, ypos, xpos) & MINEGUESS);
      else {
	if (MS_IsMine (field, ypos, xpos))
	  return MS_BLOWUP;

	field->closed -= MS_OpenUp (field, ypos, xpos);
	res = MS_OK;
      }
      break;
    case 0x1b:
    case KEY_F(0):
    case CTRL('X'):
      res = MS_EXIT;
      break;
    }
  }
  return res;
}

void MS_DisplayField (WINDOW *win, MINEFIELD *field, int ypos, int xpos, int flag)
{
  char test = MS_PeekField (field, ypos, xpos);
  int  num  = MS_FieldMines(field, ypos, xpos);

  if (flag) {
    wattrset (win, COLOR_PAIR(1));

    if (test == FREE) {
      if (num == 0)
	mvwaddch (win, ypos, xpos, ' ');
      else
	mvwaddch (win, ypos, xpos, num + '0');
    }
    else if (test & MINEGUESS)
      mvwaddch (win, ypos, xpos, 'X');
    else if (test & QUESTION)
      mvwaddch (win, ypos, xpos, '?');
    else
      mvwaddch (win, ypos, xpos, '#');
  }
  else if (test == FREE) {
    wattrset (win, COLOR_PAIR(4));

    if (num == 0)
      mvwaddch (win, ypos, xpos, ' ');
    else
      mvwaddch (win, ypos, xpos, num + '0');
  }
  else if (test & MINEGUESS) {
    wattrset (win, COLOR_PAIR(2));
    mvwaddch (win, ypos, xpos, 'X');
  }
  else if (test & QUESTION) {
    wattrset (win, COLOR_PAIR(4));
    mvwaddch (win, ypos, xpos, '?');
  }
  else {
    wattrset (win, COLOR_PAIR(3));
    mvwaddch (win, ypos, xpos, '#');
  }
}

void MS_RedrawAll (WINDOW *win, MINEFIELD *field)
{
  int ypos, xpos;
  
  for (ypos=0; ypos<field->ysize; ypos++)
    for (xpos=0; xpos<field->xsize; xpos++)
      MS_DisplayField (win, field, ypos, xpos, 0);
}

int MS_Game (WINDOW *win, int ysize, int xsize, int mines)
{
  MINEFIELD *field;
  FINFO     *infos;
  int ypos=0, xpos=0, res;
  long zeit = time (NULL);
  char dummy[10];

  if ((field = MS_InitField (ysize, xsize, mines)) == NULL)
    return;
  else if ((infos = InitForm ("MinesInfo")) == NULL) {
    MS_KillField (field);
    return;
  }

  MS_RedrawAll (win, field);
  DisplayForm  (infos);

  while (1) {
    sprintf (dummy, "%5d", field->mines - field->guess);
    PutForm (infos, 300, dummy);
    sprintf (dummy, "%5d", time (NULL) - zeit);
    PutForm (infos, 301, dummy);

    MS_DisplayField  (win, field, ypos, xpos, 1);
    res = MS_DoCursor(field, ypos, xpos);
    MS_DisplayField  (win, field, ypos, xpos, 0);

    if      (res == MS_UP)    ypos--;
    else if (res == MS_DOWN)  ypos++;
    else if (res == MS_LEFT)  xpos--;
    else if (res == MS_RIGHT) xpos++;
    else if (res == MS_OK) {
      MS_RedrawAll (win, field);

      if (field->found == field->mines && field->closed== 0)
	break;
    }
    else if (res == MS_TAG) {
      if (field->found == field->mines && field->closed == 0) {
	res = MS_OK;
	break;
      }
    }
    else if (res == MS_BLOWUP || res == MS_EXIT)
      break;
  }

  DeleteForm   (infos);
  MS_KillField (field);
  return res;
}

void Sweep (void)
{
  WINDOW *win, *sub;
  PANEL  *pan, *spa;
  int hard, ypos, xpos, res;

 nochmal:
  if ((hard = DoSbox ("MinesLevel")) == SB_ESCAPE)
    return;

  ypos = (LINES - MS_Size[hard-300][0]) / 2 + 1;
  xpos = (COLS  - MS_Size[hard-300][1]) / 2 + ((hard==302)?2:0);

  if ((win = newwin (MS_Size[hard-300][0]+4, MS_Size[hard-300][1]+6, ypos-3, xpos-3)) == NULL)
    return;
  if ((pan = new_panel (win)) == NULL) {
    delwin (win);
    return;
  }
  if ((sub = subwin (win, MS_Size[hard-300][0], MS_Size[hard-300][1], ypos, xpos)) == NULL) {
    del_panel (pan);
    delwin    (win);
    return;
  }
  if ((spa = new_panel (sub)) == NULL) {
    del_panel (pan);
    delwin    (sub);
    delwin    (win);
    return;
  }

  wattrset  (win, COLOR_PAIR (3));
  wfill     (win);
  box       (win, 0, 0);
  whorline  (win, 2);
  wattrset  (win, COLOR_PAIR (4));
  woutline  (win, "Minesweeper", 1, AL_CENTER);
  wattrset  (win, COLOR_PAIR (2));

  wbox      (win, 2, 2, MS_Size[hard-300][0] + 3, MS_Size[hard-300][1] + 3);

  SetHelpContext ("Minesweeper");

  res = MS_Game (sub,
		 MS_Size[hard-300][0],
		 MS_Size[hard-300][1],
		 MS_Size[hard-300][2]);

  if (res == MS_OK)
    InfoUser (GetMsgResource ("Prima"), GetMsgResource ("YouWon"));
  else if (res == MS_BLOWUP)
    InfoUser (GetMsgResource ("Sorry"), GetMsgResource ("BlewUp"));

  DelHelpContext ();

  res = DoSbox ("GambleAgain");

  del_panel (spa);
  del_panel (pan);
  delwin    (sub);
  delwin    (win);

  if (res == 300)
    goto nochmal;
}
